/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      CheckBox
Version:   1.0
ID:        ID_CHECKBOX
Status:    Beta
Category:  GUI
Date:      November 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems, 2003.  All rights reserved.
Keywords:  checkbox, check, box, on, off, true, false
Short:     The CheckBox class manages check boxes in the user interface.

###DESCRIPTION###
<p>The CheckBox class simplifies the creation and management of checkboxes as
part of the user interface.  Check boxes are simple gadgets that are limited
to exhibiting an on/off state.  By default, a checkbox that is turned off will
be empty, and when turned on it will display a checkmark.  The CheckBox
class allows for its graphics to be edited, so you can redefine how the on/off
states are displayed.</p>

<p>The definitions for new checkboxes are loaded by default from the
environment file "templates:checkboxdef.xml".  You can change the template file
prior to initialisation by setting the <a href="#Template">Template</a> field.  Note that any values
set in the template will override your original field settings for the
checkbox.</p>

<p>Depending on your requirements, you may need to configure the checkbox so
that when its state changes, it performs an action.  There are three acceptable
methods that you can use to achieve this.  The first is to set the
<a href="#ActionScript">ActionScript</a> field so that it refers to a script that executes when the
checkbox changes.  The second is to initialise child objects to the
checkbox and they will be executed on activation. The third option is to
listen to the Activate action by calling the <a href="../modules/kernel.html#SubscribeAction()">SubscribeAction()</a> function on
the checkbox.</p>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/files/all.h>
#include <pandora/misc/all.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/modules/render.h>
#include <pandora/input/joyport.h>
#include <pandora/input/keyboard.h>
#include <pandora/data/xml.h>
#include <pandora/functions/set.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_CHECKBOX  1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_CHECKBOX, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2003. All rights reserved.",
   "November 2003",
   "CheckBox"
};

static struct KernelBase  *KernelBase;
static struct StringsBase *StringsBase;
static struct FileSystemBase *FileSystemBase;
static struct RenderBase *RenderBase;
static OBJECTPTR CheckBoxClass = NULL;
static OBJECTPTR StringsMod = NULL;
static OBJECTPTR RenderMod = NULL;
static OBJECTPTR FileSystemMod = NULL;

enum {
   STATE_ENTERED=1,
   STATE_EXITED,
   STATE_INSIDE
};

/*****************************************************************************
** Object structure.
*/

struct CheckBox {
   OBJECT_HEADER              /* Class header */
   OBJECTID RegionID;         /* Drawable region created by the checkbox object */
   OBJECTID DrawableID;       /* The drawable target for the checkbox graphic */
   OBJECTID InputID;          /* Keyboard object to use for receiving input */
   LONG     Flags;            /* Special options */
   struct   Font *Font;       /* Font control object */
   LONG     EnterFrame;       /* Frame to display when the mouse enters the checkbox area */
   LONG     ExitFrame;        /* Frame to revert to when the mouse leaves the checkbox area */
   LONG     ClickFrame;       /* Frame to display when the checkbox is clicked */
   LONG     ReleaseFrame;     /* Frame to display when the checkbox is released */
   LONG     Thickness;        /* Border thickness */
   LONG     LabelWidth;
   LONG     Value;            /* Either TRUE or FALSE */

   PRIVATE_FIELDS
   struct Picture *Picture;   /* Bitmap image to use when the box is checked */
   struct RGB ColourRGB;      /* Colour to use as the background */
   struct RGB HighlightRGB;   /* Colour to use for border highlighting */
   struct RGB ShadowRGB;      /* Colour to use for border shadowing */
   OBJECTPTR TemplateXML;
   OBJECTID  PictureID;
   OBJECTID  FontID;
   MEMORYID  ActionScriptMID;
   STRING    ActionScript;     /* Simple DML statement indicating what to do on checkbox-press, e.g. "<set object="[desktop]" field="">"... */
   BYTE      KeyboardChannel;
   BYTE      State;
   BYTE      Active;
   UBYTE     Label[50];        /* Label to display alongside the checkbox */
};

#define CBF_HIDE       0x00000001 /* Hide on creation, or checkbox is currently hidden*/
#define CBF_DISABLED   0x00000002 /* Disable on creation */
#define CBF_NOBKGD     0x00000008 /* Do not draw background graphics (automatically set if Template graphics are used) */
#define CBF_LOCAL      0x00000010 /* On activation, execute objects locally rather than messaging them */
#define CBF_FADEBORDER 0x00000020 /* Fade the border edges into the checkbox colour (requires thickness > 1) */

/*****************************************************************************
** Class definition.
*/

static struct FieldDef CheckBoxFlags[] = {
   { "HIDE",       CBF_HIDE },
   { "DISABLED",   CBF_DISABLED },
   { "NOBKGD",     CBF_NOBKGD },
   { "LOCAL",      CBF_LOCAL },
   { "FADEBORDER", CBF_FADEBORDER },
   { NULL, NULL }
};

static ERROR GET_ActionScript(struct CheckBox *, STRING *);
static ERROR GET_Bottom(struct CheckBox *, LONG *);
static ERROR GET_ColourRGB(struct CheckBox *, struct RGB *);
static ERROR GET_Disable(struct CheckBox *, LONG *);
static ERROR GET_Height(struct CheckBox *, struct Variable *);
static ERROR GET_HighlightRGB(struct CheckBox *, struct RGB *);
static ERROR GET_Label(struct CheckBox *, STRING *);
static ERROR GET_Right(struct CheckBox *, LONG *);
static ERROR GET_ShadowRGB(struct CheckBox *, struct RGB *);
static ERROR GET_Value(struct CheckBox *, LONG *);
static ERROR GET_Width(struct CheckBox *, struct Variable *);
static ERROR GET_X(struct CheckBox *, struct Variable *);
static ERROR GET_XOffset(struct CheckBox *, struct Variable *);
static ERROR GET_Y(struct CheckBox *, struct Variable *);
static ERROR GET_YOffset(struct CheckBox *, struct Variable *);

static ERROR SET_ActionScript(struct CheckBox *, STRING);
static ERROR SET_Border(struct CheckBox *, STRING);
static ERROR SET_CheckImage(struct CheckBox *, STRING);
static ERROR SET_Colour(struct CheckBox *, STRING);
static ERROR SET_Disable(struct CheckBox *, LONG *);
static ERROR SET_Height(struct CheckBox *, struct Variable *);
static ERROR SET_Highlight(struct CheckBox *, STRING);
static ERROR SET_Label(struct CheckBox *, STRING);
static ERROR SET_Shadow(struct CheckBox *, STRING);
static ERROR SET_TabFocus(struct CheckBox *, OBJECTID *);
static ERROR SET_Template(struct CheckBox *, STRING);
static ERROR SET_Value(struct CheckBox *, LONG *);
static ERROR SET_Width(struct CheckBox *, struct Variable *);
static ERROR SET_X(struct CheckBox *, struct Variable *);
static ERROR SET_XOffset(struct CheckBox *, struct Variable *);
static ERROR SET_Y(struct CheckBox *, struct Variable *);
static ERROR SET_YOffset(struct CheckBox *, struct Variable *);

static struct FieldArray CheckBoxFields[] = {
   { "Region",         0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "Drawable",       0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "Input",          0, FDF_OBJECTID|FDF_RI,  ID_KEYBOARD, NULL ,NULL },
   { "Flags",          0, FDF_LONGFLAGS|FDF_RW, (LONG)&CheckBoxFlags, NULL, NULL },
   { "Font",           0, FDF_CHILD|FDF_R,      0, NULL, NULL },
   { "EnterFrame",     0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "ExitFrame",      0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "ClickFrame",     0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "ReleaseFrame",   0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Thickness",      0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "LabelWidth",     0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Value",          0, FDF_LONG|FDF_RW,      0, NULL, SET_Value },
   /** Virtual fields ***/
   { "ActionScript",   0, FDF_STRING|FDF_RW,   0, GET_ActionScript, SET_ActionScript },
   { "Border",         0, FDF_STRING|FDF_W,    0, NULL,             SET_Border },
   { "Bottom",         0, FDF_LONG|FDF_R,      0, GET_Bottom,       NULL },
   { "CheckImage",     0, FDF_STRING|FDF_W,    0, NULL,             SET_CheckImage },
   { "Colour",         0, FDF_STRING|FDF_W,    0, NULL,             SET_Colour },
   { "ColourRGB",      0, FDF_POINTER|FDF_R,   0, GET_ColourRGB,    NULL },
   { "Disable",        0, FDF_LONG|FDF_RW,     0, GET_Disable,      SET_Disable },
   { "Highlight",      0, FDF_STRING|FDF_W,    0, NULL,             SET_Highlight },
   { "HighlightRGB",   0, FDF_POINTER|FDF_R,   0, GET_HighlightRGB, NULL },
   { "Label",          0, FDF_STRING|FDF_RW,   0, GET_Label,        SET_Label },
   { "Right",          0, FDF_LONG|FDF_R,      0, GET_Right,        NULL },
   { "Selected",       0, FDF_LONG|FDF_RW,     0, GET_Value,        SET_Value },
   { "Shadow",         0, FDF_STRING|FDF_W,    0, NULL,             SET_Shadow },
   { "ShadowRGB",      0, FDF_POINTER|FDF_R,   0, GET_ShadowRGB,    NULL },
   { "TabFocus",       0, FDF_OBJECTID|FDF_W,  ID_TABFOCUS, NULL,   SET_TabFocus },
   { "Template",       0, FDF_STRING|FDF_I,    0, NULL,             SET_Template },
   /*** Variable Fields ***/
   { "Height",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",          0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",              0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",              0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

/*****************************************************************************
** Actions
*/

static ERROR CHECKBOX_AccessObject(struct CheckBox *, APTR);
static ERROR CHECKBOX_ActionNotify(struct CheckBox *, struct acActionNotify *);
static ERROR CHECKBOX_Activate(struct CheckBox *, APTR);
static ERROR CHECKBOX_ClosingTag(struct CheckBox *, APTR);
static ERROR CHECKBOX_DataChannel(struct CheckBox *, struct acDataChannel *);
static ERROR CHECKBOX_Disable(struct CheckBox *, APTR);
static ERROR CHECKBOX_Enable(struct CheckBox *, APTR);
static ERROR CHECKBOX_Focus(struct CheckBox *, APTR);
static ERROR CHECKBOX_Free(struct CheckBox *, APTR);
static ERROR CHECKBOX_Hide(struct CheckBox *, APTR);
static ERROR CHECKBOX_LostFocus(struct CheckBox *, APTR);
static ERROR CHECKBOX_Init(struct CheckBox *, APTR);
static ERROR CHECKBOX_MoveToBack(struct CheckBox *, APTR);
static ERROR CHECKBOX_MoveToFront(struct CheckBox *, APTR);
static ERROR CHECKBOX_NewObject(struct CheckBox *, APTR);
static ERROR CHECKBOX_Redimension(struct CheckBox *, struct acRedimension *);
static ERROR CHECKBOX_ReleaseObject(struct CheckBox *, APTR);
static ERROR CHECKBOX_Resize(struct CheckBox *, struct acResize *);
static ERROR CHECKBOX_Show(struct CheckBox *, APTR);
static ERROR CHECKBOX_UserClick(struct CheckBox *, struct acUserClick *);
static ERROR CHECKBOX_UserMovement(struct CheckBox *, struct acUserMovement *);

static struct ActionArray CheckBoxActions[] = {
   { AC_AccessObject,  CHECKBOX_AccessObject },
   { AC_ActionNotify,  CHECKBOX_ActionNotify },
   { AC_Activate,      CHECKBOX_Activate },
   { AC_ClosingTag,    CHECKBOX_ClosingTag },
   { AC_DataChannel,   CHECKBOX_DataChannel },
   { AC_Disable,       CHECKBOX_Disable },
   { AC_Enable,        CHECKBOX_Enable },
   { AC_Focus,         CHECKBOX_Focus },
   { AC_Free,          CHECKBOX_Free },
   { AC_Hide,          CHECKBOX_Hide },
   { AC_Init,          CHECKBOX_Init },
   { AC_LostFocus,     CHECKBOX_LostFocus },
   { AC_MoveToBack,    CHECKBOX_MoveToBack },
   { AC_MoveToFront,   CHECKBOX_MoveToFront },
   { AC_NewObject,     CHECKBOX_NewObject },
   { AC_Redimension,   CHECKBOX_Redimension },
   { AC_ReleaseObject, CHECKBOX_ReleaseObject },
   { AC_Resize,        CHECKBOX_Resize },
   { AC_Show,          CHECKBOX_Show },
   { AC_UserClick,     CHECKBOX_UserClick },
   { AC_UserMovement,  CHECKBOX_UserMovement },
   { NULL, NULL }
};

/*****************************************************************************
** Prototypes
*/

static void DrawCheckBox(struct CheckBox *, struct Render *, struct Bitmap *);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &StringsMod, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &FileSystemMod, NULL,
      FID_Name|TSTRING, "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(FileSystemMod, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &RenderMod, NULL,
      FID_Name|TSTRING,  "render",
      FID_Version|TFLOAT, MODVERSION_RENDER,
      TAGEND) IS ERR_Okay) {
      if (GetField(RenderMod, FID_ModBase, FT_POINTER, &RenderBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&CheckBoxClass, NULL,
      FID_BaseClassID|TLONG, ID_CHECKBOX,
      FID_Version|TFLOAT,    VER_CHECKBOX,
      FID_Name|TSTRING,      "CheckBox",
      FID_Category|TLONG,    CCF_GUI,
      FID_Flags|TLONG,       CLF_PROMOTECHILDREN,
      FID_Actions|TPTR,      CheckBoxActions,
      FID_Fields|TPTR,       CheckBoxFields,
      FID_Size|TLONG,        sizeof(struct CheckBox),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (CheckBoxClass) { Action(AC_Free, CheckBoxClass, NULL); CheckBoxClass = NULL; }
   if (StringsMod)    { Action(AC_Free, StringsMod, NULL);    StringsMod = NULL; }
   if (FileSystemMod) { Action(AC_Free, FileSystemMod, NULL); FileSystemMod = NULL; }
   if (RenderMod)     { Action(AC_Free, RenderMod, NULL);     RenderMod = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** CheckBox: AccessObject
*/

static ERROR CHECKBOX_AccessObject(struct CheckBox *Self, APTR Void)
{
   if (Self->FontID) {
      if (AccessObject(Self->FontID, 0, (OBJECTPTR *)&Self->Font) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessObject));
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************
** CheckBox: ActionNotify
*/

static ERROR CHECKBOX_ActionNotify(struct CheckBox *Self, struct acActionNotify *Args)
{
   if (Args->ActionID IS AC_UserClick) Action(AC_UserClick, Self, Args->Args);
   else if (Args->ActionID IS AC_UserMovement) Action(AC_UserMovement, Self, Args->Args);
   else if (Args->ActionID IS AC_Focus) {
      OBJECTPTR input;

      DPrintF("~Focus()","[CheckBox:%d]", Self->Head.UniqueID);

      if (Self->KeyboardChannel IS FALSE) {
         if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
            SubscribeChannel(input, Self);
            Self->KeyboardChannel = TRUE;
            ReleaseObject(input);
         }
         else ObjectError(Self, ERH_Focus, ERR_AccessObject);
      }

      DelayMsg(AC_Draw, Self->RegionID, NULL);

      StepBack();
   }
   else if (Args->ActionID IS AC_LostFocus) Action(AC_LostFocus, Self, NULL);
   else if (Args->ActionID IS AC_Disable) {
      Self->Flags |= CBF_DISABLED;
      DelayMsg(AC_Draw, Self->RegionID, NULL);
   }
   else if (Args->ActionID IS AC_Enable) {
      Self->Flags &= ~CBF_DISABLED;
      DelayMsg(AC_Draw, Self->RegionID, NULL);
   }
   else return(ERR_NoSupport);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Activates the checkbox.
###END###

*****************************************************************************/

static ERROR CHECKBOX_Activate(struct CheckBox *Self, APTR Void)
{
   struct RenderInfo info;
   STRING script;
   LONG count;
   WORD i;

   DPrintF("~Activate()","[CheckBox:%d]", Self->Head.UniqueID);

   if (Self->Active IS TRUE) {
      DPrintF("@Activate:","Warning - recursion detected");
      StepBack();
      return(ERR_Failed);
   }

   Self->Active = TRUE;

   if (GetRenderInfo(Self->RegionID, &info, sizeof(info)) IS ERR_Okay) {
      if (!(info.Flags & RNF_DISABLED)) {
         Self->Value ^= 1;

         ActionMsg(AC_Draw, Self->RegionID, NULL);

         if (GET_ActionScript(Self, &Self->ActionScript) IS ERR_Okay) {
            if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
               FID_Target|TLONG,     Self->RegionID,
               FID_Location|TSTRING, Self->ActionScript,
               TAGEND) IS ERR_Okay) {
               Action(AC_Activate, script, NULL);
               Action(AC_Free, script, NULL);
            }
         }

         if ((count = TotalChildren(Self->Head.UniqueID))) {
            struct ChildEntry list[count];
            if (ListChildren(Self->Head.UniqueID, list, &count) IS ERR_Okay) {
               if (Self->Flags & CBF_LOCAL) {
                  for (i=0; i < count; i++) ActionMsg(AC_Activate, list[i].ObjectID, NULL);
               }
               else for (i=0; i < count; i++) DelayMsg(AC_Activate, list[i].ObjectID, NULL);
            }
         }
      }
   }

   Self->Active = FALSE;

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** CheckBox: ClosingTag
*/

static ERROR CHECKBOX_ClosingTag(struct CheckBox *Self, APTR Void)
{
   if (!(Self->Flags & CBF_HIDE)) Action(AC_Show, Self, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** CheckBox: DataChannel
*/

static ERROR CHECKBOX_DataChannel(struct CheckBox *Self, struct acDataChannel *Args)
{
   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if ((Args->DataType IS DATA_KEYPRESS) AND (Args->Version IS 1)) {
      struct dcKeyEntry *key;
      WORD i;

      if ((key = (struct dcKeyEntry *)Args->Buffer)) {
         for (i=0; i < Args->TotalEntries; i++) {

            if (key[i].Flags & KQ_PRESSED) {
               if ((key[i].Value IS K_ENTER) OR (key[i].Value IS K_SPACE)) {
                  Action(AC_Activate, Self, NULL);
               }
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Disable
Short: Disables the checkbox.
###END###

*****************************************************************************/

static ERROR CHECKBOX_Disable(struct CheckBox *Self, APTR Void)
{
   /* See the ActionNotify routine to see what happens
   ** when the drawable is disabled.
   */
   DPrintF("Disable()","[CheckBox:%d]", Self->Head.UniqueID);
   ActionMsg(AC_Disable, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Enable
Short: Turns the checkbox on if it has been disabled.
###END###

*****************************************************************************/

static ERROR CHECKBOX_Enable(struct CheckBox *Self, APTR Void)
{
   /* See the ActionNotify routine to see what happens
   ** when the drawable is enabled.
   */
   DPrintF("Enable()","[CheckBox:%d]", Self->Head.UniqueID);
   ActionMsg(AC_Enable, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Focus
Short: Sets the focus on the checkbox and activates keyboard monitoring.
###END###

*****************************************************************************/

static ERROR CHECKBOX_Focus(struct CheckBox *Self, APTR Void)
{
   return(ActionMsg(AC_Focus, Self->RegionID, NULL));
}

/*****************************************************************************
** CheckBox: Free
*/

static ERROR CHECKBOX_Free(struct CheckBox *Self, APTR Void)
{
   if (Self->ActionScriptMID) {
      if (Self->ActionScript) {
         ReleaseMemory(Self->ActionScript);
         Self->ActionScript = NULL;
      }
      FreeMemoryID(Self->ActionScriptMID);
      Self->ActionScriptMID = NULL;
   }

   if (Self->PictureID) {
      if (Self->Picture) {
         Action(AC_Free, Self->Picture, NULL);
         ReleaseObject(Self->Picture);
         Self->Picture = NULL;
      }
      else ActionMsg(AC_Free, Self->PictureID, NULL);
      Self->PictureID = NULL;
   }

   if (Self->FontID) {
      if (Self->Font) {
         Action(AC_Free, Self->Font, NULL);
         ReleaseObject(Self->Font);
         Self->Font = NULL;
      }
      else ActionMsg(AC_Free, Self->FontID, NULL);
      Self->FontID = NULL;
   }

   if (Self->RegionID) {
      ActionMsg(AC_Free, Self->RegionID, NULL);
      Self->RegionID = NULL;
   }

   if (Self->TemplateXML) {
      Action(AC_Free, Self->TemplateXML, NULL);
      Self->TemplateXML = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Hide
Short: Removes the checkbox from the display.
###END###

*****************************************************************************/

static ERROR CHECKBOX_Hide(struct CheckBox *Self, APTR Void)
{
   ActionMsg(AC_Hide, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** CheckBox: Init()
*/

static ERROR CHECKBOX_Init(struct CheckBox *Self, APTR Void)
{
   struct Render *region;
   OBJECTID owner_id;

   /*** Find our parent drawable ***/

   if (!Self->DrawableID) {
      owner_id = GetOwner(Self);
      while ((owner_id) AND (GetClassID(owner_id) != ID_RENDER)) {
         owner_id = GetOwnerID(owner_id);
      }
      if (owner_id) Self->DrawableID = owner_id;
      else return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
   }

   /*** Setup keyboard ***/

   if (!Self->InputID) {
      if (FastFindObject("SystemKeyboard", NULL, &Self->InputID, 1, NULL) != ERR_Okay) {
         DPrintF("@Init:","[CheckBox] Warning: Could not find any SystemKeyboard objects.");
      }
   }

   /*** Initialise the Font ***/

   if (Action(AC_Init, Self->Font, NULL) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_Init));
   }

   /*** Calculate the width of the text label, if there is one ***/

   if ((Self->LabelWidth < 1) AND (Self->Label[0])) {
      struct mtCalcStringWidth calc;
      calc.String = Self->Label;
      calc.Length = -1;
      Action(MT_CalcStringWidth, Self->Font, &calc);
      Self->LabelWidth = calc.Width + 8;
   }

   /*** Initialise the checkbox region ***/

   if (AccessObject(Self->RegionID, 5000, (OBJECTPTR *)&region) IS ERR_Okay) {
      SetFields(region, FID_Parent|TLONG, Self->DrawableID,
                        FID_Region|TLONG, TRUE,
                        TAGEND);

      if (!(region->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if ((!(region->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y))) OR (!(region->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)))) {
            SetFields(region, FID_Height|TLONG, 18, TAGEND);
         }
      }

      if (!(region->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if ((!(region->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X))) OR (!(region->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET)))) {
            SetFields(region, FID_Width|TLONG, Self->LabelWidth + region->Height, TAGEND);
         }
      }

      if (Action(AC_Init, region, NULL) IS ERR_Okay) {
         SubscribeActionTags(region, Self,
            AC_Disable,
            AC_Enable,
            AC_Focus,
            AC_LostFocus,
            AC_UserClick,
            AC_UserMovement,
            TAGEND);
      }

      ReleaseObject(region);
   }
   else return(ERR_AccessObject);

   /*** Use the base template to create the checkbox graphics ***/

   if (Self->TemplateXML) {
      OBJECTPTR script;
      struct mtGetXMLString getxml;
      struct XMLTag **xmllist, *tag;

      if (GetFields(Self->TemplateXML, FID_Tags|TPTR, &xmllist, TAGEND) IS ERR_Okay) {
         for (tag=xmllist[0]; tag; tag=tag->Next) {
            if (tag->Child) {
               if (StrCompare("graphics", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  getxml.Index = tag->Child->Index;
                  getxml.Flags = XMF_INCLUDESIBLINGS;
                  Action(MT_GetXMLString, Self->TemplateXML, &getxml);

                  if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
                     FID_String|TSTRING, getxml.Result,
                     FID_Target|TLONG,   Self->RegionID,
                     TAGEND) IS ERR_Okay) {

                     if (acActivate(script) IS ERR_Okay) {
                        Self->Flags |= CBF_NOBKGD;
                     }

                     Action(AC_Free, script, NULL);
                  }

                  FreeMemory(getxml.Result);
                  break;
               }
            }
         }
      }

      Action(AC_Free, Self->TemplateXML, NULL);
      Self->TemplateXML = NULL;
   }

   if (AccessObject(Self->RegionID, 5000, (OBJECTPTR *)&region) IS ERR_Okay) {
      ActionTags(MT_SetRenderCallback, region, Self, &DrawCheckBox);
      ReleaseObject(region);
   }
   else return(ERR_AccessObject);

   /*** Load the picture to be displayed when the checkbox is ticked ***/

   if (!Self->Picture) {
      SET_CheckImage(Self, "templates:images/tick.png");
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  LostFocus
Short: Removes the focus from the checkbox and disables key monitoring.
###END###

*****************************************************************************/

static ERROR CHECKBOX_LostFocus(struct CheckBox *Self, APTR Void)
{
   OBJECTPTR input;

   if (Self->KeyboardChannel IS TRUE) {
      if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
         UnsubscribeChannel(input, Self->Head.UniqueID);
         Self->KeyboardChannel = FALSE;
         ReleaseObject(input);
      }
      else ObjectError(Self, ERH_LostFocus, ERR_AccessObject);
   }

   DelayMsg(AC_Draw, Self->RegionID, NULL);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  MoveToBack
Short: Moves the checkbox to the back of the display area.
###END###

*****************************************************************************/

static ERROR CHECKBOX_MoveToBack(struct CheckBox *Self, APTR Void)
{
   ActionMsg(AC_MoveToBack, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  MoveToFront
Short: Moves the checkbox to the front of the display area.
###END###

*****************************************************************************/

static ERROR CHECKBOX_MoveToFront(struct CheckBox *Self, APTR Void)
{
   ActionMsg(AC_MoveToFront, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** CheckBox: NewObject()
*/

static ERROR CHECKBOX_NewObject(struct CheckBox *Self, APTR Void)
{
   if (NewObject(ID_RENDER, Self->Head.ObjectFlags|NF_CHILD, NULL, &Self->RegionID) IS ERR_Okay) {
      if (NewObject(ID_FONT, Self->Head.ObjectFlags|NF_CHILD, &Self->Font, &Self->FontID) IS ERR_Okay) {
         SetField(Self->Font, FID_Face, FT_STRING, "sans serif:9");

         Self->ExitFrame = 1;
         Self->ReleaseFrame = 1;

         /*** Shadow colour ***/
         Self->ShadowRGB.Red   = 100;
         Self->ShadowRGB.Green = 100;
         Self->ShadowRGB.Blue  = 100;

         /*** Internal colour ***/
         Self->ColourRGB.Red   = 255;
         Self->ColourRGB.Green = 255;
         Self->ColourRGB.Blue  = 255;

         /*** Highlight colour ***/
         Self->HighlightRGB.Red   = 255;
         Self->HighlightRGB.Green = 255;
         Self->HighlightRGB.Blue  = 255;

         Self->Thickness = 1;

         SET_Template(Self, "templates:checkboxdef.xml");

         return(ERR_Okay);
      }
      else return(ERR_NewObject);
   }
   else return(ERR_NewObject);
}

/*****************************************************************************

###ACTION###
Name:  Redimension
Short: Changes the size and position of the checkbox.
###END###

*****************************************************************************/

static ERROR CHECKBOX_Redimension(struct CheckBox *Self, struct acRedimension *Args)
{
   return(ActionMsg(AC_Redimension, Self->RegionID, Args));
}

/*****************************************************************************
** CheckBox: ReleaseObject
*/

static ERROR CHECKBOX_ReleaseObject(struct CheckBox *Self, APTR Void)
{
   if ((Self->Font) AND (Self->FontID)) {
      ReleaseObject(Self->Font);
      Self->Font = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Resize
Short: Alters the size of the checkbox.
###END###

*****************************************************************************/

static ERROR CHECKBOX_Resize(struct CheckBox *Self, struct acResize *Args)
{
   return(ActionMsg(AC_Resize, Self->RegionID, Args));
}

/*****************************************************************************

###ACTION###
Name:  Show
Short: Puts the checkbox on display.
###END###

*****************************************************************************/

static ERROR CHECKBOX_Show(struct CheckBox *Self, APTR Void)
{
   ActionMsg(AC_Show, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  UserClick
Short: Emulates a user click on the checkbox.
###END###

*****************************************************************************/

static ERROR CHECKBOX_UserClick(struct CheckBox *Self, struct acUserClick *Args)
{
   OBJECTPTR drawable;

   if (!Args) return(ObjectError(Self, ERH_UserClick, ERR_Args));

   if (Self->Flags & CBF_DISABLED) return(ERR_Okay);

   if (Args->XCoord < Self->LabelWidth) return(ERR_Okay);

   if (Args->ButtonFlags & JD_LMB) {
      if (Self->ClickFrame) {
         if (AccessObject(Self->RegionID, 3000, &drawable) IS ERR_Okay) {
            SetField(drawable, FID_Frame, FT_LONG, &Self->ClickFrame);
            ReleaseObject(drawable);
         }
      }

      Action(AC_Activate, Self, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  UserMovement
Short: Emulates user movement over the checkbox.
###END###

*****************************************************************************/

static ERROR CHECKBOX_UserMovement(struct CheckBox *Self, struct acUserMovement *Args)
{
   OBJECTPTR drawable;

   if (!Args) return(ObjectError(Self, ERH_UserMovement, ERR_Args));

   if (Args->OverObject IS Self->RegionID) {
      if (Self->State IS STATE_ENTERED) Self->State = STATE_INSIDE;
      else if (Self->State != STATE_INSIDE) Self->State = STATE_ENTERED;
   }
   else {
      if (Self->State IS STATE_EXITED) return(ERR_Okay);
      else Self->State = STATE_EXITED;
   }

   /*** Change the drawable's frame if necessary ***/

   if (Self->State != STATE_INSIDE) {
      if (AccessObject(Self->RegionID, 2000, (OBJECTPTR *)&drawable) IS ERR_Okay) {
         if (Self->State IS STATE_EXITED) {
            SetField(drawable, FID_Frame, FT_LONG, &Self->ExitFrame);
            DelayMsg(AC_Draw, Self->RegionID, NULL);
         }
         else if (Self->State IS STATE_ENTERED) {
            if (!(Self->Flags & CBF_DISABLED)) {
               SetField(drawable, FID_Frame, FT_LONG, &Self->EnterFrame);
               DelayMsg(AC_Draw, Self->RegionID, NULL);
            }
         }
         ReleaseObject(drawable);
      }
   }

   if (Self->State IS STATE_ENTERED) Self->State = STATE_INSIDE;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ActionScript
Short:  Script to be executed when the checkbox is clicked.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>When a checkbox is clicked, you can elect to run a script that performs a
responsive action.  The ActionScript string will typically refer to a file
location, although the special STRING: assignment followed with script
execution text (e.g. DML) can also be useful for optimising execution for
simple scripts.</p>

###END###

*****************************************************************************/

static ERROR GET_ActionScript(struct CheckBox *Self, STRING *Value)
{
   if (Self->ActionScript) {
      *Value = Self->ActionScript;
      return(ERR_Okay);
   }
   else if (!Self->ActionScriptMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->ActionScriptMID, MEM_READ, (void **)&Self->ActionScript) IS ERR_Okay) {
      *Value = Self->ActionScript;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError((OBJECTPTR)Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_ActionScript(struct CheckBox *Self, STRING Value)
{
   LONG i;

   if (Self->ActionScript)    { ReleaseMemory(Self->ActionScript);   Self->ActionScript = NULL; }
   if (Self->ActionScriptMID) { FreeMemoryID(Self->ActionScriptMID); Self->ActionScriptMID = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->ActionScript, &Self->ActionScriptMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->ActionScript[i] = Value[i];
         Self->ActionScript[i] = 0;
      }
      else return(ObjectError((OBJECTPTR)Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Border
Short:  String-based field for setting a single-colour border for the checkbox.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The border colour for a checkbox can be declared by writing to this field.
The colour must be in hexadecimal or separated-decimal format - for example
to create a pure red colour, a setting of "#ff0000" or "255,0,0" would be
valid.</p>

###SEE ALSO###
Fields: Highlight, Shadow, Thickness

###END###

*****************************************************************************/

static ERROR SET_Border(struct CheckBox *Self, STRING Colour)
{
   if (Colour) {
      StrReadColour(Colour, &Self->HighlightRGB.Red, &Self->HighlightRGB.Green, &Self->HighlightRGB.Blue, &Self->HighlightRGB.Alpha);
      Self->ShadowRGB.Red   = Self->HighlightRGB.Red;
      Self->ShadowRGB.Green = Self->HighlightRGB.Green;
      Self->ShadowRGB.Blue  = Self->HighlightRGB.Blue;
      Self->ShadowRGB.Alpha = Self->HighlightRGB.Alpha;
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Bottom
Short:  The bottom coordinate of the checkbox.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>The bottom coordinate of the checkbox (calculated as YCoord + Height) is
readable from this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Bottom(struct CheckBox *Self, LONG *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Bottom, FT_LONG, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
}

/*****************************************************************************

###FIELD###
Name:   CheckImage
Short:  Defines the image that is displayed when the checkbox state is set to 'on'.
Type:   LONG
Status: Set

###DESCRIPTION###
<p>When the checkbox state is set to 'on', it will display an image inside
the checkbox area.  You can redefine the default image by providing the path
to an image file in this field.  If the image file cannot be loaded for
any reason, the checkbox will revert to its default image on
initialisation.</p>

###END###

*****************************************************************************/

static ERROR SET_CheckImage(struct CheckBox *Self, STRING Value)
{
   if (CreateObject(ID_PICTURE, Self->Head.ObjectFlags|NF_CHILD, &Self->Picture, &Self->PictureID,
      FID_Location|TSTRING, Value,
      FID_Flags|TLONG,      PCF_32BITALPHA|PCF_CACHE,
      TAGEND) IS ERR_Okay) {

      return(ERR_Okay);
   }
   else return(ERR_CreateObject);
}

/*****************************************************************************

###FIELD###
Name:   Colour
Short:  String-based field for setting the colour inside of the checkbox.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The colour inside of a checkbox is declared by writing to this field.
The colour must be in hexadecimal or separated-decimal format - for example
to create a pure red colour, a setting of "#ff0000" or "255,0,0" would be
valid.</p>

<p>If you create a checkbox without setting the Colour field, the checkbox will
be drawn in black.</p>

###SEE ALSO###
Fields: Highlight, Shadow, Thickness

###END###

*****************************************************************************/

static ERROR SET_Colour(struct CheckBox *Self, STRING Colour)
{
   if (Colour) {
      StrReadColour(Colour, &Self->ColourRGB.Red, &Self->ColourRGB.Green, &Self->ColourRGB.Blue, &Self->ColourRGB.Alpha);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ClickFrame
Short:  The graphics frame to display when the checkbox is clicked.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the drawable frame to switch to when the user
clicks on the checkbox.  By default this field is initially set to
zero, which has no effect on the drawable frame.  When the user releases
the checkbox, it will revert to the frame indicated by the ReleaseFrame
field.</p>

###SEE ALSO###
Field: ReleaseFrame

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ColourRGB
Short:  Defines the internal colour of the checkbox in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The colour inside of the checkbox can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_ColourRGB(struct CheckBox *Self, struct RGB *Value)
{
   Value->Red   = Self->ColourRGB.Red;
   Value->Green = Self->ColourRGB.Green;
   Value->Blue  = Self->ColourRGB.Blue;
   Value->Alpha = Self->ColourRGB.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Disable
Short:  Disables the checkbox on initalisation.
Type:   BOOLEAN
Status: Get/Set

###DESCRIPTION###
<p>The checkbox can be disabled on initialisation by setting this field to TRUE.
If you need to disable the combobox after it has been activated, it is preferred
that you use the Disable action.</p>

<p>To enable the combobox after it has been disabled, use the Enable action.</p>

###END###

*****************************************************************************/

static ERROR GET_Disable(struct CheckBox *Self, LONG *Value)
{
   if (Self->Flags & CBF_DISABLED) *Value = TRUE;
   else *Value = FALSE;
   return(ERR_Okay);
}

static ERROR SET_Disable(struct CheckBox *Self, LONG *Value)
{
   if (*Value IS TRUE) Action(AC_Disable, Self, NULL);
   else if (*Value IS FALSE) Action(AC_Enable, Self, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  The drawable that will contain the checkbox graphic.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>The drawable that will contain the checkbox graphic is set here.  If this
field is not set prior to initialisation, the checkbox will attempt to scan
for the correct drawable by analysing its parents until it finds a suitable
candidate.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   EnterFrame
Short:  The graphics frame to display when the user's pointer enters the checkbox area.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the drawable frame to switch to when the user's
pointer enters the checkbox area.  By default this field is initially set to
zero, which has no effect on the drawable frame.</p>

###SEE ALSO###
Field: ExitFrame

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ExitFrame
Short:  The graphics frame to display when the user's pointer leaves the checkbox area.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the drawable frame to switch to when the user's pointer
leaves the checkbox area.  This field is unused if the EnterFrame field has
not been set.</p>

###SEE ALSO###
Field: EnterFrame

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Optional flags may be set here.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Optional flags that may be set against a checkbox object are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">DISABLED</td><td>This readable flag is automatically set when the checkbox is disabled.  You may set it on initialisation if the checkbox is to be disabled on creation.</td></tr>
<tr><td width="10%">HIDE</td><td>This flag may be set on initialisation in order to hide the checkbox on its creation (applies to script usage only).</td></tr>
<tr><td width="10%">LOCAL</td><td>By default, the checkbox will send messages to its child objects when the checkbox state changes.  If you set the LOCAL option, the checkbox will execute the objects within its own process space.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Height
Short:  Defines the height of a checkbox.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A checkbox can be given a fixed or relative height by setting this field
to the desired value.  To set a relative height, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Height(struct CheckBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetHeight, ERR_AccessObject));
}

static ERROR SET_Height(struct CheckBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (((Value->Type & FD_DOUBLE) AND (!Value->Double)) OR
       ((Value->Type & FD_LARGE) AND (!Value->Large))) {
      return(ERR_Okay);
   }

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_Height, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   Highlight
Short:  String-based field for setting the checkbox highlight.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The border highlight of a checkbox is set by writing to this field. The colour
must be in hexadecimal or separated-decimal format - for example to create a
pure red colour, a setting of "#ff0000" or "255,0,0" would be valid.</p>

###SEE ALSO###
Fields: Colour, Shadow, Thickness

###END###

*****************************************************************************/

static ERROR SET_Highlight(struct CheckBox *Self, STRING Value)
{
   if (Value) {
      StrReadColour(Value, &Self->HighlightRGB.Red, &Self->HighlightRGB.Green, &Self->HighlightRGB.Blue, &Self->HighlightRGB.Alpha);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   HighlightRGB
Short:  Defines border highlight of the checkbox, in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The border highlight of the checkbox can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_HighlightRGB(struct CheckBox *Self, struct RGB *Value)
{
   Value->Red   = Self->HighlightRGB.Red;
   Value->Green = Self->HighlightRGB.Green;
   Value->Blue  = Self->HighlightRGB.Blue;
   Value->Alpha = Self->HighlightRGB.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Label
Short:  The label is a string displayed to the left of the input area.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>A label can be drawn next to the input area by setting the Label field.
The label should be a short, descriptive string of one or two words.  It
is common practice for the label to be followed with a colon character.</p>

###END###

*****************************************************************************/

static ERROR GET_Label(struct CheckBox *Self, STRING *Value)
{
   *Value = Self->Label;
   return(ERR_Okay);
}

static ERROR SET_Label(struct CheckBox *Self, STRING Value)
{
   WORD i;
   if (Value) {
      for (i=0; (Value[i]) AND (i < sizeof(Self->Label)-1); i++) Self->Label[i] = Value[i];
      Self->Label[i] = 0;
   }
   else Self->Label[0] = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Region
Short:  The drawable that represents the checkbox is referenced through this field.
Type:   OBJECTID
Status: Read

###DESCRIPTION###
<p>The rendering area that represents the checkbox display can be accessed
through this field.  For further information, refer to the Render class.  Note
that talking to the drawable directly can have adverse effects on the checkbox
control system.  Where possible, all communication should be limited to the
checkbox object itself.</p>

###SEE ALSO###
Class: Render

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ReleaseFrame
Short:  The graphics frame to display when a user-click is released.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If the ClickFrame field has been set, you may want to match that value
by indicating the frame that should be used when the click is released.
By default, the value in this field will initially be set to 1.  This field
is unused if the ClickFrame field has not been set.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Right
Short:  The right coordinate of the checkbox.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>The right coordinate of the checkbox (calculated as XCoord + Width) is
readable from this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Right(struct CheckBox *Self, LONG *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Right, FT_LONG, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
}

/*****************************************************************************

###FIELD###
Name:   Shadow
Short:  String-based field for setting the checkbox shadow.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The border shadow of a checkbox is set by writing to this field. The colour
must be in hexadecimal or separated-decimal format - for example to create a
pure red colour, a setting of "#ff0000" or "255,0,0" would be valid.</p>

###SEE ALSO###
Fields: Colour, Highlight, Thickness

###END###

*****************************************************************************/

static ERROR SET_Shadow(struct CheckBox *Self, STRING Value)
{
   if (Value) {
      StrReadColour(Value, &Self->ShadowRGB.Red, &Self->ShadowRGB.Green, &Self->ShadowRGB.Blue, &Self->ShadowRGB.Alpha);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ShadowRGB
Short:  Defines border shadow of the checkbox, in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The border shadow of the checkbox can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_ShadowRGB(struct CheckBox *Self, struct RGB *Value)
{
   Value->Red   = Self->ShadowRGB.Red;
   Value->Green = Self->ShadowRGB.Green;
   Value->Blue  = Self->ShadowRGB.Blue;
   Value->Alpha = Self->ShadowRGB.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   TabFocus
Short:  Set this field to a TabFocus object to register the checkbox in a tab-list.
Type:   OBJECTID
Status: Set

###DESCRIPTION###
<p>The TabFocus field provides a convenient way of adding the checkbox to a
TabFocus object, so that it can be focussed on via the tab key.  Simply set
this field to the ID of the TabFocus object that is managing the tab-list
for the application window.</p>

###SEE ALSO###
Class: TabFocus

###END###

*****************************************************************************/

static ERROR SET_TabFocus(struct CheckBox *Self, OBJECTID *Value)
{
   OBJECTPTR tabfocus;

   if (AccessObject(*Value, 5000, &tabfocus) IS ERR_Okay) {
      if (tabfocus->ClassID IS ID_TABFOCUS) {
         ActionTags(MT_AddTabObject, tabfocus, Self->RegionID);
      }
      ReleaseObject(tabfocus);
   }
   else return(ERR_AccessObject);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Template
Short:  Defines the makeup of the checkbox using a pre-defined template.
Type:   STRING
Status: Init

###DESCRIPTION###
<p>Setting the Template field allows a checkbox defintion to be loaded from a
pre-defined template file.  The template may contain settings for all of the
field values of a checkbox object, as well as a customised graphics script
for building the look and feel of the checkbox.  The template must be provided
as an XML file that can contain two possible root tags - 'value' and 'graphics'.
The value tag contains a list of field values for setting against the checkbox,
while the graphics tag contains the graphical makeup of the checkbox.  The
following example illustrates how a customised checkbox can be created:</p>

<pre>
&lt;?xml version="1.0"?&gt;

&lt;values&gt;
  &lt;border value="#000000"/&gt;
&lt;/values&gt;

&lt;graphics&gt;
  &lt;box boxes="(r0,0,!0,!0) (r1,1,!1,!1)" colour="230,230,230"
    highlight(0)="120,120,120" highlight(1)="255,255,255" shadow(0)="80,80,80"
    shadow(1)="200,200,200"/&gt;
&lt;/graphics&gt;
</pre>

###END###

*****************************************************************************/

static ERROR SET_Template(struct CheckBox *Self, STRING Value)
{
   struct XMLTag **xmllist, *tag, *child;
   UBYTE value[300];
   WORD i;

   if (Self->TemplateXML) {
      Action(AC_Free, Self->TemplateXML, NULL);
      Self->TemplateXML = NULL;
   }

   if (AnalyseLocation(Value, NULL) IS ERR_Okay) {
      if (CreateObject(ID_XML, NF_CHILD, &Self->TemplateXML, NULL,
         FID_Location|TSTRING, Value,
         TAGEND) IS ERR_Okay) {

         if (GetFields(Self->TemplateXML, FID_Tags|TPTR, &xmllist, TAGEND) IS ERR_Okay) {
            for (tag=xmllist[0]; tag; tag=tag->Next) {
               if (tag->Child) {
                  if (StrCompare("values", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     for (child=tag->Child; child; child=child->Next) {
                        if (child->TotalAttributes > 1) {
                           for (i=0; (i < sizeof(value)-1) AND (child->Attributes[1].Value[i]); i++) value[i] = child->Attributes[1].Value[i];
                           value[i] = 0;
                           StrTranslate(value, sizeof(value), NULL);
                           SetFieldVariable(Self, child->Attributes[0].Name, value);
                        }
                     }
                     break;
                  }
               }
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Value
Short:  Indicates the current TRUE/FALSE status for the checkbox.
Type:   BOOLEAN
Status: Read/Set

###DESCRIPTION###
<p>To get the TRUE/FALSE status of the checkbox, read this field.  You can
also set this field if you wish to change the state of the checkbox.  Only
values of 0 and 1 are valid.</p>

###END###

*****************************************************************************/

static ERROR GET_Value(struct CheckBox *Self, LONG *Value)
{
   *Value = Self->Value;
   return(ERR_Okay);
}

static ERROR SET_Value(struct CheckBox *Self, LONG *Value)
{
   DPrintF("SetField:","[CheckBox.Value] = %d", *Value);

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      if ((*Value IS TRUE) AND (Self->Value != TRUE)) {
         Self->Value = TRUE;
         ActionMsg(AC_Draw, Self->RegionID, NULL);
      }
      else if ((*Value IS FALSE) AND (Self->Value != FALSE)) {
         Self->Value = FALSE;
         ActionMsg(AC_Draw, Self->RegionID, NULL);
      }
   }
   else Self->Value = *Value;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  Defines the width of a checkbox.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A checkbox can be given a fixed or relative width by setting this field
to the desired value.  To set a relative width, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Width(struct CheckBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetWidth, ERR_AccessObject));
}

static ERROR SET_Width(struct CheckBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (((Value->Type & FD_DOUBLE) AND (!Value->Double)) OR
       ((Value->Type & FD_LARGE) AND (!Value->Large))) {
      return(ERR_Okay);
   }

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_Width, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  The horizontal position of a checkbox.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The horizontal position of a checkbox can be set to an absolute or relative
coordinate by writing a value to the XCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct CheckBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_XCoord, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetX, ERR_AccessObject));
}

static ERROR SET_X(struct CheckBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_XCoord, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  The horizontal offset of a checkbox.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the checkbox will be
drawn from that X coordinate up to the width of the container, minus the
value given in the XOffset.  This means that the width of the CheckBox is
dynamically calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the checkbox will be drawn at an X coordinate calculated from the formula
&quot;XCoord = ContainerWidth - CheckBoxWidth - XOffset&quot;.</p>

###SEE ALSO###
Field: XCoord, Width, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct CheckBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_XOffset, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetXOffset, ERR_AccessObject));
}

static ERROR SET_XOffset(struct CheckBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_XOffset, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  The vertical position of a checkbox.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The vertical position of a CheckBox can be set to an absolute or relative
coordinate by writing a value to the YCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: YOffset, XCoord

###END###

*****************************************************************************/

static ERROR GET_Y(struct CheckBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_YCoord, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetY, ERR_AccessObject));

}

static ERROR SET_Y(struct CheckBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_YCoord, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  The vertical offset of a checkbox.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the checkbox will be drawn
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the checkbox is
dynamically calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the checkbox will be drawn at a Y coordinate calculated from the formula
&quot;YCoord = ContainerHeight - CheckBoxHeight - YOffset&quot;.</p>

###SEE ALSO###
Field: YCoord, Height, XOffset

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct CheckBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_YOffset, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetYOffset, ERR_AccessObject));
}

static ERROR SET_YOffset(struct CheckBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_YOffset, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}


/*****************************************************************************
** Internal: DrawCheckBox()
*/

static void DrawCheckBox(struct CheckBox *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   struct mtDrawRGBRectangle rect;
   struct Font *font;
   ULONG highlight, shadow;
   WORD i;

   if (!(Self->Flags & CBF_NOBKGD)) {
      rect.XCoord = Self->LabelWidth;
      rect.YCoord = 0;
      rect.Width  = Drawable->Height;
      rect.Height = Drawable->Height;
      rect.Fill   = TRUE;
      rect.Red    = Self->ColourRGB.Red;
      rect.Green  = Self->ColourRGB.Green;
      rect.Blue   = Self->ColourRGB.Blue;
      if ((Self->State IS STATE_INSIDE) OR (Self->State IS STATE_ENTERED)) {
         if (!(Drawable->Flags & RNF_DISABLED)) {
            if ((rect.Red   += 20) > 255) rect.Red   = 255;
            if ((rect.Green += 20) > 255) rect.Green = 255;
            if ((rect.Blue  += 20) > 255) rect.Blue  = 255;
         }
      }
      Action(MT_DrawRGBRectangle, Bitmap, &rect);

      shadow    = PackPixel(Bitmap, Self->ShadowRGB.Red, Self->ShadowRGB.Green, Self->ShadowRGB.Blue);
      highlight = PackPixel(Bitmap, Self->HighlightRGB.Red, Self->HighlightRGB.Green, Self->HighlightRGB.Blue);

      i = 0;
      if (Self->Thickness > 0) {
         for (i=0; i < Self->Thickness; i++) {
            /*** Top, Bottom ***/

            ActionTags(MT_DrawRectangle, Bitmap, Self->LabelWidth + i, i, Drawable->Height-i-i, 1, shadow, TRUE);
            ActionTags(MT_DrawRectangle, Bitmap, Self->LabelWidth + i, Drawable->Height-i-1, Drawable->Height-i-i, 1, highlight, TRUE);

            /*** Left, Right ***/

            ActionTags(MT_DrawRectangle, Bitmap, Self->LabelWidth + i,                   i+1, 1, Drawable->Height-i-i-2, shadow, TRUE);
            ActionTags(MT_DrawRectangle, Bitmap, Self->LabelWidth + Drawable->Height-i-1, i+1, 1, Drawable->Height-i-i-2, highlight, TRUE);
         }
      }

      if (Drawable->Flags & RNF_HASFOCUS) {
         if (!(Drawable->Flags & RNF_DISABLED)) {
            Bitmap->Opacity = 32;
            ActionTags(MT_DrawRGBRectangle, Bitmap, Self->LabelWidth + i, i, Drawable->Height-i-i, Drawable->Height-i-i, 0, 0, 0, FALSE);
            Bitmap->Opacity = 255;
         }
      }
   }

   if (Self->Label[0]) {
      font = Self->Font;

      SetField(font, FID_Bitmap, FT_POINTER, Bitmap);
      SetField(font, FID_String, FT_STRING, Self->Label);

      if (Drawable->Flags & RNF_DISABLED) SetFields(font, FID_Opacity|TLONG, 25, TAGEND);

      font->XCoord = 0;
      font->YCoord = 0;
      font->AlignFlags |= ALIGN_VERTICAL;
      font->AlignWidth  = Drawable->Width;
      font->AlignHeight = Drawable->Height;
      Action(AC_Draw, font, NULL);

      if (Drawable->Flags & RNF_DISABLED) SetFields(font, FID_Opacity|TLONG, 100, TAGEND);
   }

   if (Self->Picture) {
      struct Bitmap *srcbitmap;

      srcbitmap = Self->Picture->Bitmap;

      if (srcbitmap->Flags & BMF_ALPHA) srcbitmap->Flags |= BMF_ALPHABLEND;

      if (Self->Value IS FALSE) srcbitmap->Opacity = 25;

      ActionTags(MT_BlitArea, srcbitmap, Bitmap, NULL, 0, 0, srcbitmap->Width, srcbitmap->Height,
         Self->LabelWidth + ((Drawable->Height - srcbitmap->Width)/2), ((Drawable->Height - srcbitmap->Height)/2));

      srcbitmap->Opacity = 255;
      srcbitmap->Flags &= ~BMF_ALPHABLEND;
   }
}

