/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Window
Version:   1.0
ID:        ID_WINDOW
Status:    Beta
Category:  GUI
Date:      July 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems, 2003-2004.  All rights reserved.
Keywords:  window, layer
Short:     The Window class is used to create desktop windows and screen areas.

###DESCRIPTION###
<p>The Window class provides a control system for the creation and
management of application windows within the user interface.  Because the
Window class acts as a control system only, the window itself is virtual and
physically represented as a Render drawable.  The characteristics of the
<a href="render.html">Render</a> class are inherited by the window system, thereby allowing you to
read and manipulate render fields (such as x, y, width and height) through
the window object.</p>

<p>By default, new windows are created in the desktop area with the
dimensions that you specify.  If desirable, you can create full-screen
windows by setting the <a href="#Screen">Screen</a> field to TRUE.  Full-screen windows have a
different set of specifications that keep them maximised to the display area
at all times.  They are borderless but still maintain a title bar and
optional menu bar if required.  If your application creates a full-screen
window then all future windows will open within that screen space by default
rather than on the desktop.</p>

<p>Window definitions are loaded by default from the environment file
"templates:windowdef.xml".  You can change the template file prior to
initialisation by setting the <a href="#Template">Template</a> field, however this is not
recommended except under exceptional circumstances.  See the <a href="#Template">Template</a> field
for details on how to design your own windows for custom environments.</p>

<p>Windows are publicly shared objects as standard and may not be allocated
privately. This is automatically managed when using windows in scripts.  If
programming a window object directly, remember to grab its unique ID and
access/release the window object correctly as you make use of it.</p>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/files/all.h>
#include <pandora/misc/all.h>
#include <pandora/input/joyport.h>
#include <pandora/data/script.h>
#include <pandora/data/xml.h>
#include <pandora/functions/set.c>

#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/modules/render.h>
#include <pandora/modules/screen.h>

#undef NULL
#define NULL 0

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_WINDOW  1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_WINDOW, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2003-2004. All rights reserved.",
   "July 2004",
   "Window"
};

static struct KernelBase *KernelBase;
static struct StringsBase *StringsBase;
static struct RenderBase *RenderBase;
static struct FileSystemBase *FileSystemBase;
static struct ScreenBase *ScreenBase;
static OBJECTPTR WindowClass = NULL;
static OBJECTPTR modStrings = NULL;
static OBJECTPTR modFileSystem = NULL;
static OBJECTPTR modRender = NULL;
static OBJECTPTR modScreen = NULL;
static OBJECTID glDefaultScreen = NULL;

/*****************************************************************************
** Class definition.
*/

static struct FieldDef WindowFlags[] = {
   { "DISABLED",    WNF_DISABLED },
   { "SMARTLIMITS", WNF_SMARTLIMITS },
   { "MANUALCLOSE", WNF_MANUALCLOSE },
   { "BACKGROUND",  WNF_BACKGROUND },
   { "VIDEO",       WNF_VIDEO },
   { "NOMARGINS",   WNF_NOMARGINS },
   { NULL, NULL }
};

static struct FieldDef ResizeFlags[] = {
   { "TOP",         EDGE_TOP },
   { "LEFT",        EDGE_LEFT },
   { "RIGHT",       EDGE_RIGHT },
   { "BOTTOM",      EDGE_BOTTOM },
   { "TOPLEFT",     EDGE_TOPLEFT },
   { "TOPRIGHT",    EDGE_TOPRIGHT },
   { "BOTTOMLEFT",  EDGE_BOTTOMLEFT },
   { "BOTTOMRIGHT", EDGE_BOTTOMRIGHT },
   { "ALL",         EDGE_ALL },
   { NULL, NULL }
};

static ERROR GET_Canvas(struct Window *, OBJECTID *);
static ERROR GET_Icon(struct Window *, STRING *);
static ERROR GET_Menu(struct Window *, STRING *);
static ERROR GET_ScriptArgs(struct Window *, STRING *);
static ERROR GET_Title(struct Window *, STRING *);

static ERROR SET_Close(struct Window *, LONG *);
static ERROR SET_Icon(struct Window *, STRING);
static ERROR SET_Maximise(struct Window *, LONG *);
static ERROR SET_Menu(struct Window *, STRING);
static ERROR SET_Minimise(struct Window *, LONG *);
static ERROR SET_MoveToBack(struct Window *, LONG *);
static ERROR SET_Screen(struct Window *, LONG *);
static ERROR SET_StickToFront(struct Window *, LONG *);
static ERROR SET_Template(struct Window *, STRING);
static ERROR SET_Title(struct Window *, STRING);

static struct FieldArray WindowFields[] = {
   { "Drawable",         0, FDF_CHILD|FDF_R,      ID_RENDER,          NULL, NULL },
   { "Flags",            0, FDF_LONGFLAGS|FDF_RW, (LONG)&WindowFlags, NULL, NULL },
   { "InsideBorder",     0, FDF_LONG|FDF_RI,      0,                  NULL, NULL },
   { "Center",           0, FDF_LONG|FDF_RI,      0,                  NULL, NULL },
   { "Minimise",         0, FDF_LONG|FDF_RW,      0,                  NULL, SET_Minimise },
   { "Maximise",         0, FDF_LONG|FDF_RW,      0,                  NULL, SET_Maximise },
   { "MoveToBack",       0, FDF_LONG|FDF_RW,      0,                  NULL, SET_MoveToBack },
   { "Close",            0, FDF_LONG|FDF_RW,      0,                  NULL, SET_Close },
   { "Quit",             0, FDF_LONG|FDF_RW,      0,                  NULL, NULL },
   { "RestoreX",         0, FDF_LONG|FDF_RW,      0,                  NULL, NULL },
   { "RestoreY",         0, FDF_LONG|FDF_RW,      0,                  NULL, NULL },
   { "RestoreWidth",     0, FDF_LONG|FDF_RW,      0,                  NULL, NULL },
   { "RestoreHeight",    0, FDF_LONG|FDF_RW,      0,                  NULL, NULL },
   { "Hide",             0, FDF_LONG|FDF_RW,      0,                  NULL, NULL },
   { "Focus",            0, FDF_LONG|FDF_RI,      0,                  NULL, NULL },
   { "Screen",           0, FDF_LONG|FDF_RW,      0,                  NULL, SET_Screen },
   { "TitleObject",      0, FDF_OBJECTID|FDF_RI,  0,                  NULL, NULL },
   { "MinimiseObject",   0, FDF_OBJECTID|FDF_RI,  0,                  NULL, NULL },
   { "MaximiseObject",   0, FDF_OBJECTID|FDF_RI,  0,                  NULL, NULL },
   { "MoveToBackObject", 0, FDF_OBJECTID|FDF_RI,  0,                  NULL, NULL },
   { "CloseObject",      0, FDF_OBJECTID|FDF_RI,  0,                  NULL, NULL },
   { "MenubarObject",    0, FDF_OBJECTID|FDF_RI,  0,                  NULL, NULL },
   { "ScreenMargin",     0, FDF_LONG|FDF_RI,      0,                  NULL, NULL },
   { "Resize",           0, FDF_LONGFLAGS|FDF_I,  (LONG)&ResizeFlags, NULL, NULL },
   { "ResizeBorder",     0, FDF_LONG|FDF_RI,      0,                  NULL, NULL },
   { "NoMargins",        0, FDF_LONG|FDF_I,       0,                  NULL, NULL },
   { "Canvas",           0, FDF_OBJECTID|FDF_R,   0,                  GET_Canvas, NULL },
   /** Virtual fields ***/
   { "Icon",           0, FDF_STRING|FDF_RW, 0, GET_Icon,       SET_Icon },
   { "Menu",           0, FDF_STRING|FDF_RW, 0, GET_Menu,       SET_Menu },
   { "ScriptArgs",     0, FDF_STRING|FDF_RW, 0, GET_ScriptArgs, NULL },
   { "StickToFront",   0, FDF_LONG|FDF_W,    0, NULL,           SET_StickToFront },
   { "Template",       0, FDF_STRING|FDF_I,  0, NULL,           SET_Template },
   { "Title",          0, FDF_STRING|FDF_RW, 0, GET_Title,      SET_Title },
   END_FIELD
};

/*****************************************************************************
** Actions
*/

static ERROR WINDOW_AccessObject(struct Window *, APTR);
static ERROR WINDOW_ActionNotify(struct Window *, struct acActionNotify *);
static ERROR WINDOW_Activate(struct Window *, APTR);
static ERROR WINDOW_ClosingTag(struct Window *, APTR);
static ERROR WINDOW_Disable(struct Window *, APTR);
static ERROR WINDOW_Enable(struct Window *, APTR);
static ERROR WINDOW_Focus(struct Window *, APTR);
static ERROR WINDOW_Free(struct Window *, APTR);
static ERROR WINDOW_GetUnlistedField(struct Window *, struct acGetUnlistedField *);
static ERROR WINDOW_Hide(struct Window *, APTR);
static ERROR WINDOW_Init(struct Window *, APTR);
static ERROR WINDOW_Move(struct Window *, struct acMove *);
static ERROR WINDOW_MoveToPoint(struct Window *, struct acMoveToPoint *);
static ERROR WINDOW_MoveToBack(struct Window *, APTR);
static ERROR WINDOW_MoveToFront(struct Window *, APTR);
static ERROR WINDOW_NewChild(struct Window *, struct acNewChild *);
static ERROR WINDOW_NewObject(struct Window *, APTR);
static ERROR WINDOW_NewOwner(struct Window *, struct acNewOwner *);
static ERROR WINDOW_Redimension(struct Window *, struct acRedimension *);
static ERROR WINDOW_ReleaseObject(struct Window *, APTR);
static ERROR WINDOW_Resize(struct Window *, struct acResize *);
static ERROR WINDOW_SetUnlistedField(struct Window *, struct acSetUnlistedField *);
static ERROR WINDOW_Show(struct Window *, APTR);

static struct ActionArray WindowActions[] = {
   { AC_AccessObject,     WINDOW_AccessObject },
   { AC_ActionNotify,     WINDOW_ActionNotify },
   { AC_Activate,         WINDOW_Activate },
   { AC_ClosingTag,       WINDOW_ClosingTag },
   { AC_Disable,          WINDOW_Disable },
   { AC_Enable,           WINDOW_Enable },
   { AC_Focus,            WINDOW_Focus },
   { AC_Free,             WINDOW_Free },
   { AC_GetUnlistedField, WINDOW_GetUnlistedField },
   { AC_Hide,             WINDOW_Hide },
   { AC_Init,             WINDOW_Init },
   { AC_Move,             WINDOW_Move },
   { AC_MoveToPoint,      WINDOW_MoveToPoint },
   { AC_MoveToBack,       WINDOW_MoveToBack },
   { AC_MoveToFront,      WINDOW_MoveToFront },
   { AC_NewChild,         WINDOW_NewChild },
   { AC_NewObject,        WINDOW_NewObject },
   { AC_NewOwner,         WINDOW_NewOwner },
   { AC_Redimension,      WINDOW_Redimension },
   { AC_ReleaseObject,    WINDOW_ReleaseObject },
   { AC_Resize,           WINDOW_Resize },
   { AC_SetUnlistedField, WINDOW_SetUnlistedField },
   { AC_Show,             WINDOW_Show },
   { NULL, NULL }
};

/*****************************************************************************
** Class methods.
*/

static ERROR WINDOW_Close(struct Window *, APTR);
static ERROR WINDOW_Maximise(struct Window *, struct mtMaximiseWindow *);
static ERROR WINDOW_Minimise(struct Window *, APTR);

static struct FunctionField argsMaximiseWindow[] = { { "Toggle", ARG_LONG }, { NULL, NULL } };

static struct MethodArray WindowMethods[] = {
   { MT_MaximiseWindow, WINDOW_Maximise, "Maximise", argsMaximiseWindow, sizeof(struct mtMaximiseWindow) },
   { MT_MinimiseWindow, WINDOW_Minimise, "Minimise", NULL, NULL },
   { MT_CloseWindow,    WINDOW_Close,    "Close",    NULL, NULL },
   { NULL, NULL, NULL, NULL }
};

/*****************************************************************************
** Prototypes
*/

struct VarString {
   STRING Field;
   STRING Value;
};

static void CheckWindowOverlap(struct Window *);
static struct XMLTag * GetXMLTemplate(struct Window *, STRING);
static STRING GetXMLTemplateString(struct Window *, STRING);
static void ProcessArguments(struct Window *, OBJECTPTR);
static void SmartLimits(struct Window *);
static void set_variables(OBJECTPTR, struct VarString *);

static struct VarString varColours[] = {
   { "shadow",     "80,80,80" },
   { "highlight",  "255,255,255" },
   { "colour",     "200,200,200" },
   { "button",     "200,200,200" },
   { "menubar",    "190,190,190" },
   { "menu",       "240,240,240" },
   { "menuborder", "0,0,0" },
   { "white",      "255,255,255" },
   { "red",        "255,0,0" },
   { "green",      "0,255,0" },
   { "blue",       "0,0,255" },
   { "background", "255,255,255" },
   { "desktop",    "160,160,160" },
   { "recess",     "128,128,128" },
   { NULL, NULL }
};

static struct VarString varFonts[] = {
   { "face",            "sans serif:9"},
   { "colour",          "255,255,255"},
   { "height",          "11"},
   { "window_face",     "sans serif:9"},
   { "window_colour",   "#000000"},
   { "window_height",   "11"},
   { "button_face",     "sans serif:9"},
   { "button_colour",   "#000000"},
   { "button_height",   "11"},
   { "icon_face",       "sans serif:9"},
   { "icon_colour",     "#ffffff"},
   { "icon_height",     "11"},
   { "small_face",      "sans serif:9"},
   { "small_colour",    "#ffffff"},
   { "small_height",    "11"},
   { "menu_face",       "sans serif:9"},
   { "menu_colour",     "#000000"},
   { "menu_highlight",  "#ffffff"},
   { "menu_height",     "11"},
   { "titlebar_face",   "sans serif:9"},
   { "titlebar_colour", "#ffffff"},
   { "titlebar_height", "11"},
   { "large_face",      "sans serif:9"},
   { "large_colour",    "#ffffff"},
   { "large_height",    "11"},
   { NULL, NULL }
};

static struct VarString varInterface[] = {
   { "buttonheight",    "21"},
   { "gadgetgap",       "6"},
   { "textborder",      "1"},
   { "inputborder",     "1"},
   { "inputborder_x",   "1"},
   { "inputborder_y",   "1"},
   { "inputborder_xo",  "0"},
   { "inputborder_yo",  "0"},
   { "outputborder",    "1"},
   { "outputborder_x",  "1"},
   { "outputborder_y",  "1"},
   { "outputborder_xo", "0"},
   { "outputborder_yo", "0"},
   { "slidersize",      "16"},
   { "scrollsize",      "18"},
   { NULL, NULL }
};

static struct VarString varDesktop[] = {
   { "wallpaper",         "pictures:textures/fishnet.jpg"},
   { "wallpaper_style",   "Tile"},
   { "iconratio",         "7.5"},
   { "icontheme",         "Default"},
   { "iconsize",          "0"},
   { "colour",            "58,110,165"},
   { "screensaver",       "programs:multimedia/screensavers/Starfield"},
   { "screensaver_delay", "300"},
   { "monitoroff_delay",  "1200"},
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   OBJECTPTR object;
   OBJECTID object_id;

   KernelBase = argKernelBase;

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modRender, NULL,
      FID_Name|TSTRING,   "render",
      FID_Version|TFLOAT, MODVERSION_RENDER,
      TAGEND) IS ERR_Okay) {
      if (GetField(modRender, FID_ModBase, FT_POINTER, &RenderBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modScreen, NULL,
      FID_Name|TSTRING,   "screen",
      FID_Version|TFLOAT, MODVERSION_SCREEN,
      TAGEND) IS ERR_Okay) {
      if (GetField(modScreen, FID_ModBase, FT_POINTER, &ScreenBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   /* Initialise default variables if they do not exist already.  Note that in
   ** Pandora, scripts are not available, so we have to hardwire the default
   ** variables.
   */

   if (CheckObjectExists(NULL, "gl_colours") != ERR_Okay) {
      if (CreateObject(ID_SCRIPT, NULL, &object, NULL,
            FID_Location|TSTRING, "templates:defaultvariables",
            TAGEND) IS ERR_Okay) {
         acActivate(object);
         acFree(object);
      }

      if (CheckObjectExists(NULL, "gl_colours") != ERR_Okay) {
         if (CreateObject(ID_VARIABLE, NF_PUBLIC|NF_SHARED|NF_UNIQUE|NF_UNTRACKED, &object, &object_id,
               FID_Name|TSTRING, "gl_colours",
               TAGEND) IS ERR_Okay) {
            set_variables(object, varColours);
            ReleaseObject(object);
         }

         if (CreateObject(ID_VARIABLE, NF_PUBLIC|NF_SHARED|NF_UNIQUE|NF_UNTRACKED, &object, &object_id,
               FID_Name|TSTRING, "gl_fonts",
               TAGEND) IS ERR_Okay) {
            set_variables(object, varFonts);
            ReleaseObject(object);
         }

         if (CreateObject(ID_VARIABLE, NF_PUBLIC|NF_SHARED|NF_UNIQUE|NF_UNTRACKED, &object, &object_id,
               FID_Name|TSTRING, "gl_interface",
               TAGEND) IS ERR_Okay) {
            set_variables(object, varInterface);
            ReleaseObject(object);
         }

         if (CreateObject(ID_VARIABLE, NF_PUBLIC|NF_SHARED|NF_UNIQUE|NF_UNTRACKED, &object, &object_id,
               FID_Name|TSTRING, "gl_desktop",
               TAGEND) IS ERR_Okay) {
            set_variables(object, varDesktop);
            ReleaseObject(object);
         }
      }
   }

   if (CheckObjectExists(NULL, "SystemKeyboard") != ERR_Okay) {
      if (CreateObject(ID_KEYBOARD, NF_PUBLIC|NF_UNTRACKED|NF_UNIQUE, &object, &object_id,
            FID_Name|TSTRING, "SystemKeyboard",
            TAGEND) IS ERR_Okay) {
         acActivate(object);
         ReleaseObject(object);
      }
   }

   if (NewNamedObject(ID_ICONSERVER, NF_NOTRACK|NF_PUBLIC|NF_UNIQUE,
      &object, &object_id, "SystemIcons") IS ERR_Okay) {
      if (acInit(object) != ERR_Okay) acFree(object);
      acActivate(object);
      ReleaseObject(object);
   }

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&WindowClass, NULL,
      FID_BaseClassID|TLONG, ID_WINDOW,
      FID_Version|TFLOAT,    VER_WINDOW,
      FID_Name|TSTRING,      "Window",
      FID_Category|TLONG,    CCF_GUI,
      FID_Flags|TLONG,       CLF_PROMOTECHILDREN|CLF_SHAREDOBJECTS|CLF_PUBLICOBJECTS,
      FID_Actions|TPTR,      WindowActions,
      FID_Methods|TPTR,      WindowMethods,
      FID_Fields|TPTR,       WindowFields,
      FID_Size|TLONG,        sizeof(struct Window),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (WindowClass)   { acFree(WindowClass);   WindowClass   = NULL; }
   if (modStrings)    { acFree(modStrings);    modStrings    = NULL; }
   if (modFileSystem) { acFree(modFileSystem); modFileSystem = NULL; }
   if (modRender)     { acFree(modRender);     modRender     = NULL; }
   if (modScreen)     { acFree(modScreen);     modScreen     = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Window: AccessObject
*/

static ERROR WINDOW_AccessObject(struct Window *Self, APTR Void)
{
   if ((Self->DrawableID) AND (!Self->Drawable)) {
      if (AccessObject(Self->DrawableID, 4000, (OBJECTPTR *)&Self->Drawable) != ERR_Okay) {
         DPrintF("AccessObject:","[Window:%d] Failed to access drawable #%d.", Self->Head.UniqueID, Self->DrawableID);
         Self->Drawable = NULL;
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Window: ActionNotify
*/

static ERROR WINDOW_ActionNotify(struct Window *Self, struct acActionNotify *Args)
{
   if (Args->ActionID IS AC_Focus) {
      if (!(Self->Head.ObjectFlags & NF_INITIALISED)) return(ERR_Okay);

      DPrintF("~Focus()","[Window:%d]", Self->Head.UniqueID);

      /*** Move the window to the front when the focus is received ***/

      if (Self->Screen IS FALSE) {
         acMoveToFront(Self);
      }

      /*** Ensure that the window is visible when the focus is received ***/

      if ((!(Self->Drawable->Flags & RNF_VISIBLE)) AND (Self->Shown)) Action(AC_Show, Self, NULL);

      StepBack();

      NotifySubscribers(Self, AC_Focus, NULL, NULL);
   }
   else if (Args->ActionID IS AC_Disable) {
      Self->Flags |= WNF_DISABLED;
      DelayMsg(AC_Draw, Self->DrawableID, NULL);
   }
   else if (Args->ActionID IS AC_Enable) {
      Self->Flags &= ~WNF_DISABLED;
      DelayMsg(AC_Draw, Self->DrawableID, NULL);
   }
   else if (Args->ActionID IS AC_Free) {
      if (Args->ObjectID IS Self->DrawableID) {
         Action(AC_Free, Self, NULL);
      }
   }
   else if (Args->ActionID IS AC_LostFocus) {
      Action(AC_LostFocus, Self, NULL);
      NotifySubscribers(Self, AC_LostFocus, NULL, NULL);
   }
   else if (Args->ActionID IS AC_Redimension) {
      SmartLimits(Self);
   }
   else return(ERR_NoSupport);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Shows the window.
###END###

*****************************************************************************/

static ERROR WINDOW_Activate(struct Window *Self, APTR Void)
{
   acShow(Self);
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    CloseWindow
Synonym: Close
Short:   Closes the window according to application requirements.

###DESCRIPTION###
<p>When terminating a window, it is recommended that you call the CloseWindow() method,
as opposed to forcibly removing it with the Free action.  By using the Close
method, the application will have the opportunity to respond in a way that
is appropriate for the window in question.  When the Close method is typically
called, the application will terminate entirely if the Quit field has been set
to a value of TRUE.  If this option is not set then the window will terminate
quietly without affecting the core application.</p>

<p>If you would like to write a customised routine for window closure, set
the MANUALCLOSE flag for the window and then subscribe to the Close method (or
set the Quit field and wait for a quit message on the message queue).
This will allow you to receive close notifications and you can respond in a
way that suits the current situation.</p>

###END###

*****************************************************************************/

static ERROR WINDOW_Close(struct Window *Self, APTR Void)
{
   if (Self->Close IS FALSE) return(ERR_Okay); /* Developer has requested that closing the window not be possible */

   /* If WNF_MANUALCLOSE is set then it's up to the developer to use the
   ** Free action and kill the window manually.
   */

   if (!(Self->Flags & WNF_MANUALCLOSE)) {
      DPrintF("~Close()","[Window:%d]", Self->Head.UniqueID);
      acFree(Self);
      StepBack();
   }
   else {
      DPrintF("Close()","[Window:%d] The program will kill the window manually.", Self->Head.UniqueID);

      /* WHen MANUALCLOSE is used, the developer can either opt to listen
      ** to the CLose method, or alternatively he can set the Quit field
      ** to TRUE and we will send him a QUIT message.
      */

      if (Self->Quit) {
         DPrintF("Close:","[Window] Sending the application a quit message.");

         if (Self->Head.TaskID IS CurrentTaskID()) {
            SendMessage(NULL, MSGID_QUIT, NULL, NULL, NULL);
         }
         else {
            struct ListTasks *list;
            OBJECTID task_id;
            WORD i, message;

            task_id = CurrentTaskID();
            message = FALSE;
            if (ListTasks(NULL, &list) IS ERR_Okay) {
               for (i=0; list[i].TaskID; i++) {
                  if (list[i].TaskID IS task_id) {
                     SendMessage(list[i].MessageID, MSGID_QUIT, NULL, NULL, NULL);
                     message = TRUE;
                     break;
                  }
               }
               FreeMemory(list);
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Window: ClosingTag
*/

static ERROR WINDOW_ClosingTag(struct Window *Self, APTR Void)
{
   struct RenderInfo info;
   struct acMoveToPoint moveto;

   DPrintF("~ClosingTag()","[Window:%d]", Self->Head.UniqueID);

   /* Recalculate the window center, just in case the dimensions of the window
   ** were modified between the opening and closing tag statements.
   */

   if ((Self->Center IS TRUE) AND (Self->Screen IS FALSE)) {
      if (Self->Drawable->ParentID) {
         if (GetRenderInfo(Self->Drawable->ParentID, &info, sizeof(info)) IS ERR_Okay) {
            moveto.XCoord = (info.Width - Self->Drawable->Width) / 2;
            moveto.YCoord = (info.Height - Self->Drawable->Height) / 2;
            moveto.Flags  = MTF_X|MTF_Y;
            if ((moveto.XCoord != Self->Drawable->XCoord) OR (moveto.YCoord != Self->Drawable->YCoord)) {
               Action(AC_MoveToPoint, Self->Drawable, &moveto);
               CheckWindowOverlap(Self);
            }
         }
      }
   }

   if (Self->Hide IS FALSE) acShow(Self);

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Disable
Short: Disables the user's ability to interact with the window.
###END###

*****************************************************************************/

static ERROR WINDOW_Disable(struct Window *Self, APTR Void)
{
   /* See the ActionNotify routine to see what happens
   ** when the drawable is disabled.
   */

   DPrintF("Disable()","[Window:%d]", Self->Head.UniqueID);

   acDisable(Self->Drawable);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Enable
Short: Enables user interactivity after prior disablement.
###END###

*****************************************************************************/

static ERROR WINDOW_Enable(struct Window *Self, APTR Void)
{
   /* See the ActionNotify routine to see what happens
   ** when the drawable is enabled.
   */
   DPrintF("Enable()","[Window:%d]", Self->Head.UniqueID);
   acEnable(Self->Drawable);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Focus
Short: Sets the user focus to the window's drawable.
###END###

*****************************************************************************/

static ERROR WINDOW_Focus(struct Window *Self, APTR Void)
{
   /*** See our ActionNotify routine for notification support ***/
   return(Action(AC_Focus, Self->Drawable, NULL)|ERF_Notified);
}

/*****************************************************************************
** Window: Free
*/

static ERROR WINDOW_Free(struct Window *Self, APTR Void)
{
   WORD i;

   acHide(Self);

   if (Self->Drawable) {
      UnsubscribeAction(Self->Drawable, NULL, Self->Head.UniqueID);
   }

   if (Self->DrawableID) {
      if (Self->Drawable) {
         acFree(Self->Drawable);
         ReleaseObject(Self->Drawable);
         Self->Drawable = NULL;
      }
      else acFreeID(Self->DrawableID);
      Self->DrawableID = NULL;
   }

   for (i=0; i < Self->TotalArgs; i++) {
      if (Self->LocalArgs[i]) FreeMemory(Self->LocalArgs[i]);
   }

   if (Self->LocalArgs) {
      FreeMemory(Self->LocalArgs);
      Self->LocalArgs = NULL;
   }

   if (Self->ScriptArgs) {
       FreeMemory(Self->ScriptArgs);
      Self->ScriptArgs = NULL;
   }

   if (Self->TemplateXML) {
      Action(AC_Free, Self->TemplateXML, NULL);
      Self->TemplateXML = NULL;
   }

   if (Self->Quit IS TRUE) {
      DPrintF("Free:","[Window] Sending the application a quit message.");

      if (Self->Head.TaskID IS CurrentTaskID()) {
         SendMessage(NULL, MSGID_QUIT, NULL, NULL, NULL);
      }
      else {
         struct ListTasks *list;
         OBJECTID task_id;
         WORD i, message;

         task_id = CurrentTaskID();
         message = FALSE;
         if (ListTasks(NULL, &list) IS ERR_Okay) {
            for (i=0; list[i].TaskID; i++) {
               if (list[i].TaskID IS task_id) {
                  SendMessage(list[i].MessageID, MSGID_QUIT, NULL, NULL, NULL);
                  message = TRUE;
                  break;
               }
            }
            FreeMemory(list);
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  GetUnlistedField
Short: Pass-through arguments can be retrieved through this action.
###END###

*****************************************************************************/

static ERROR WINDOW_GetUnlistedField(struct Window *Self, struct acGetUnlistedField *Args)
{
   STRING arg;
   WORD i;

   for (i=0; i < Self->TotalArgs; i++) {
      if ((arg = Self->LocalArgs[i])) {
         if (StrCompare(Args->Field, arg, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            while (*arg) arg++;
            arg++;
            StrCopy(arg, Args->Buffer, Args->Size);
            return(ERR_Okay);
         }
      }
   }

   return(ERR_UnsupportedField);
}

/*****************************************************************************

###ACTION###
Name:  Hide
Short: Removes the window from the display.
###END###

*****************************************************************************/

static ERROR WINDOW_Hide(struct Window *Self, APTR Void)
{
   struct RenderList *list;
   OBJECTID parent_id, window_id;
   WORD i;
   LONG size;

   if (!Self->Drawable) return(ERR_Okay); /* Sometimes there is no drawable if this routine is called from Free() */

   if (Self->Drawable->Flags & RNF_HASFOCUS) {
      DPrintF("~Hide()","[Window:%d]", Self->Head.UniqueID);

      acHide(Self->Drawable);

      /*** Find the top-most window in our container and change the focus to it ***/

      if ((parent_id = Self->Drawable->ParentID)) {
         if ((list = AccessRenderList(ARF_READ, &size))) {
            for (i=0; list->ObjectID; i++, list = ((APTR)list) + size);

            for (--i; (i >= 0); i--, list = ((APTR)list) - size) {
               if ((list->ParentID IS parent_id) AND (list->ObjectID != Self->Drawable->Head.UniqueID)) {
                  if (list->Flags & RNF_VISIBLE) {
                     if ((window_id = GetOwnerID(list->ObjectID))) {
                        if (GetClassID(window_id) IS ID_WINDOW) {
                           acFocusID(window_id);
                           break;
                        }
                     }
                  }
               }
            }

            ReleaseRenderList(ARF_READ);
         }
         else DPrintF("@Hide:","[Window] Access to the renderlist denied.");
      }

      StepBack();
   }
   else acHide(Self->Drawable);

   return(ERR_Okay);
}

/*****************************************************************************
** Window: Init()
*/

static ERROR WINDOW_Init(struct Window *Self, APTR Void)
{
   struct acMoveToPoint moveto;
   struct RenderInfo info;
   OBJECTPTR script, xml;
   STRING str;

   if (!Self->Drawable->ParentID) {
      /* Screen mode not applicable if there is no parent */
      Self->Screen = FALSE;

      if ((ScreenDisplayType() IS DT_NATIVE) OR (Self->Flags & WNF_NOMARGINS)) {
         /*** Turn off margins when in fullscreen mode ***/

         Self->Drawable->Width  -= Self->Drawable->LeftMargin + Self->Drawable->RightMargin; /* Get a correct 'insidewidth' */
         Self->Drawable->Height -= Self->Drawable->TopMargin + Self->Drawable->BottomMargin; /* Get a correct 'insideheight' */
         Self->Drawable->LeftMargin   = 0;
         Self->Drawable->TopMargin    = 0;
         Self->Drawable->RightMargin  = 0;
         Self->Drawable->BottomMargin = 0;
         Self->Screen = FALSE; /* Turn off the draggable screen option because we don't want to use that template */
      }
      else {
         /*** When in hosted mode, force default window margins ***/

         Self->Drawable->LeftMargin   = 6;
         Self->Drawable->TopMargin    = 6;
         Self->Drawable->RightMargin  = 6;
         Self->Drawable->BottomMargin = 6;
      }

      /* Allow video render buffers when in full screen mode */
      if (Self->Flags & WNF_VIDEO) Self->Drawable->Flags |= RNF_VIDEO;
   }
   else {
      /*** If the NOMARGINS flag is used, adjust the margins to expectations ***/

      if ((Self->Flags & WNF_NOMARGINS) AND (Self->NoMargins)) {
         if (Self->NoMargins < 0) Self->NoMargins = -Self->NoMargins;

         Self->Drawable->LeftMargin   -= Self->NoMargins;
         Self->Drawable->TopMargin    -= Self->NoMargins;
         Self->Drawable->RightMargin  -= Self->NoMargins;
         Self->Drawable->BottomMargin -= Self->NoMargins;

         SetFields(Self->Drawable, FID_Width|TLONG,  Self->Drawable->Width  - (Self->NoMargins * 2),
                                   FID_Height|TLONG, Self->Drawable->Height - (Self->NoMargins * 2),
                                   TAGEND);
      }
   }

   if ((Self->Drawable->ParentID) AND (Self->Screen IS FALSE)) Self->Drawable->Flags |= RNF_ROOTLAYER|RNF_PERVASIVECOPY;

   /*** In draggable screen mode, ensure that the dimensions of the window cover the entire display ***/

   if (Self->Screen IS TRUE) {
      SetFields(Self->Drawable,
         FID_MaxHeight|TLONG, 4096,
         FID_MaxWidth|TLONG,  4096,
         FID_MinHeight|TLONG, 40,
         FID_MinWidth|TLONG,  40,
         FID_XCoord|TLONG,    0,
         FID_YCoord|TLONG,    0,
         FID_Height|TLONG,    0,
         FID_Width|TLONG,     0,
         FID_XOffset|TLONG,   0,
         FID_YOffset|TLONG,   0,
         TAGEND);
      /* Allow video render buffers when in draggable screen mode */
      if (Self->Flags & WNF_VIDEO) Self->Drawable->Flags |= RNF_VIDEO;
   }

   DPrintF("Init:","[Window] Dimensions: %dx%d,%dx%d, Margins: %d,%d,%d,%d", Self->Drawable->XCoord, Self->Drawable->YCoord, Self->Drawable->Width, Self->Drawable->Height, Self->Drawable->LeftMargin, Self->Drawable->TopMargin, Self->Drawable->RightMargin, Self->Drawable->BottomMargin);

   if (acInit(Self->Drawable) IS ERR_Okay) {
      SubscribeActionTags(Self->Drawable, Self,
         AC_Disable,
         AC_Enable,
         AC_Focus,
         AC_Free,
         AC_LostFocus,
         AC_Redimension,
         TAGEND);
   }

   /*** Turn off the maximise gadget if the maximium and minimum values are equal ***/

   if ((Self->Drawable->MaxHeight IS Self->Drawable->MinHeight) AND
       (Self->Drawable->MaxWidth IS Self->Drawable->MinWidth)) {
      Self->Maximise = FALSE;
   }

   /*** Move the window to the center of the display if centering is turned on ***/

   if ((Self->Center IS TRUE) AND (Self->Screen IS FALSE) AND (Self->Drawable->ParentID)) {
      if (GetRenderInfo(Self->Drawable->ParentID, &info, sizeof(info)) IS ERR_Okay) {
         moveto.XCoord = (info.Width - Self->Drawable->Width) / 2;
         moveto.YCoord = (info.Height - Self->Drawable->Height) / 2;
         moveto.Flags  = MTF_X|MTF_Y;
         Action(AC_MoveToPoint, Self->Drawable, &moveto);
      }
   }

   /* Check if our window position conflicts with that of other render objects
   ** within our parent region.  If so, adjust the position so that we are slightly
   ** offset.
   */

   CheckWindowOverlap(Self);

   /*** Ensure that the window is within the display area ***/

   if (Self->Drawable->ParentID) {
      if (GetRenderInfo(Self->Drawable->ParentID, &info, sizeof(info)) IS ERR_Okay) {
         /*** Check position against limits ***/

         if (Self->Drawable->XCoord + Self->Drawable->Width > info.Width - Self->Drawable->RightLimit) {
            SetFields(Self->Drawable, FID_XCoord|TLONG, (LONG)(info.Width - Self->Drawable->RightLimit - Self->Drawable->Width), TAGEND);
         }

         if (Self->Drawable->YCoord + Self->Drawable->Height > info.Height - Self->Drawable->BottomLimit) {
            SetFields(Self->Drawable, FID_YCoord|TLONG, info.Height - Self->Drawable->BottomLimit - Self->Drawable->Height, TAGEND);
         }

         if (Self->Drawable->XCoord < Self->Drawable->LeftLimit) {
            SetFields(Self->Drawable, FID_XCoord|TLONG, Self->Drawable->LeftLimit, TAGEND);
         }

         if (Self->Drawable->YCoord < Self->Drawable->TopLimit) {
            SetFields(Self->Drawable, FID_YCoord|TLONG, Self->Drawable->TopLimit, TAGEND);
         }

         /*** Check position against basic width/height dimensions of the parent ***/

         if (Self->Drawable->XCoord + Self->Drawable->Width > info.Width) {
            SetFields(Self->Drawable, FID_XCoord|TLONG, (LONG)(info.Width - Self->Drawable->Width), TAGEND);
         }

         if (Self->Drawable->YCoord + Self->Drawable->Height > info.Height) {
            SetFields(Self->Drawable, FID_YCoord|TLONG, info.Height - Self->Drawable->Height, TAGEND);
         }

         if (Self->Drawable->XCoord < 0) {
            SetFields(Self->Drawable, FID_XCoord|TLONG, 0, TAGEND);
         }

         if (Self->Drawable->YCoord < 0) {
            SetFields(Self->Drawable, FID_YCoord|TLONG, 0, TAGEND);
         }
      }
   }

   /*** Use the base template to create the window graphics ***/

   if (Self->Drawable->ParentID) {

      if (!Self->TemplateXML) {
         DPrintF("@Init:","[Window] Window XML template has not been defined.");
         return(ERR_FieldNotSet);
      }

      /*** Run the graphics script ***/

      if ((str = GetXMLTemplateString(Self, "graphics"))) {
         if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
            FID_String|TSTRING, str,
            FID_Target|TLONG,   Self->DrawableID,
            FID_Flags|TLONG,    SCF_EXITONERROR,
            FID_Window|TUNLISTED|TLONG, Self->Head.UniqueID,
            TAGEND) IS ERR_Okay) {

            if (acActivate(script) != ERR_Okay) {
               acFree(script);
               DPrintF("@Init:","[Window] Failed to generate window template graphics.");
               return(ERR_Activate);
            }

            acFree(script);
         }
         else {
            DPrintF("@Init:","[Window] Failed to create window script.");
            return(ERR_CreateObject);
         }

         FreeMemory(str);
      }
      else {
         DPrintF("@Init:","[Window] Window graphics specification missing from XML file.");
         return(ERR_Failed);
      }

      /*** Create the title bar ***/

      if ((str = GetXMLTemplateString(Self, "titlebar"))) {
         if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
            FID_String|TSTRING,  str,
            FID_Target|TLONG,    Self->DrawableID,
            FID_Flags|TLONG,     SCF_EXITONERROR,
            FID_Title|TUNLISTED, Self->Title,
            FID_Window|TUNLISTED|TLONG, Self->Head.UniqueID,
            TAGEND) IS ERR_Okay) {

            if (acActivate(script) != ERR_Okay) {
               acFree(script);
               DPrintF("@Init:","[Window] Failed to generate window titlebar graphics.");
               return(ERR_Activate);
            }

            acFree(script);
         }
         else {
            DPrintF("@Init:","[Window] Failed to create window script.");
            return(ERR_CreateObject);
         }

         FreeMemory(str);
      }
   }
   else if (ScreenDisplayType() IS DT_NATIVE) {
      /* Do not create background graphics when in full screen
      ** mode, unless WNF_BACKGROUND is set, which indicates
      ** that a background is absolutely required.
      */

      if (Self->Flags & WNF_BACKGROUND) {
         if (CreateObject(ID_BOX, NULL, NULL, NULL,
            FID_Owner|TLONG,   Self->DrawableID,
            FID_XCoord|TLONG,  0,
            FID_YCoord|TLONG,  0,
            FID_XOffset|TLONG, 0,
            FID_YOffset|TLONG, 0,
            FID_Colour|TTRANS, "[gl_colours.colour]",
            TAGEND) != ERR_Okay) {
            return(ObjectError(Self, ERH_Init, ERR_CreateObject));
         }
      }
   }
   else {
      if (CreateObject(ID_BOX, NULL, NULL, NULL,
         FID_Owner|TLONG,   Self->DrawableID,
         FID_XCoord|TLONG,  0,
         FID_YCoord|TLONG,  0,
         FID_XOffset|TLONG, 0,
         FID_YOffset|TLONG, 0,
         FID_Colour|TTRANS, "[gl_colours.colour]",
         TAGEND) != ERR_Okay) {
         return(ObjectError(Self, ERH_Init, ERR_CreateObject));
      }

      if (Self->InsideBorder IS TRUE) {
         if (CreateObject(ID_BOX, NULL, NULL, NULL,
            FID_Owner|TLONG,       Self->DrawableID,
            FID_XCoord|TLONG,      Self->Drawable->LeftMargin - 1,
            FID_YCoord|TLONG,      Self->Drawable->TopMargin - 1,
            FID_XOffset|TLONG,     Self->Drawable->RightMargin - 1,
            FID_YOffset|TLONG,     Self->Drawable->BottomMargin - 1,
            FID_Highlight|TSTRING, "#dfdfdf",
            FID_Shadow|TSTRING,    "#202020",
            FID_Sunken|TLONG,      TRUE,
            TAGEND) != ERR_Okay) {
            return(ObjectError(Self, ERH_Init, ERR_CreateObject));
         }

         if (CreateObject(ID_BOX, NULL, NULL, NULL,
            FID_Owner|TLONG,       Self->DrawableID,
            FID_XCoord|TLONG,      Self->Drawable->LeftMargin - 2,
            FID_YCoord|TLONG,      Self->Drawable->TopMargin - 2,
            FID_XOffset|TLONG,     Self->Drawable->RightMargin - 2,
            FID_YOffset|TLONG,     Self->Drawable->BottomMargin - 2,
            FID_Highlight|TSTRING, "#ffffff",
            FID_Shadow|TSTRING,    "#808080",
            FID_Sunken|TLONG,      TRUE,
            TAGEND) != ERR_Okay) {
            return(ObjectError(Self, ERH_Init, ERR_CreateObject));
         }
      }
   }

   /* Create a menubar if one has been requested.  The user will have provided
   ** the location of the menubar XML file in the Menu field.  We will call
   ** the menubar script as specified in our window definition template.
   */

   if (Self->Menu[0]) {
      DPrintF("Init:","[Window] Processing menu file.");

      if ((str = GetXMLTemplateString(Self, "menubar"))) {

         /*** Load the XML menu file ***/

         if (CreateObject(ID_XML, NF_CHILD, &xml, NULL,
            FID_Location|TSTRING, Self->Menu,
            TAGEND) IS ERR_Okay) {

            if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
               FID_String|TSTRING, str,
               FID_Target|TLONG,   Self->DrawableID,
               FID_Window|TUNLISTED|TLONG, Self->Head.UniqueID,
               FID_XML|TUNLISTED|TLONG,    xml->UniqueID,
               TAGEND) IS ERR_Okay) {

               ProcessArguments(Self, script);

               acActivate(script);
               acFree(script);
            }

            acFree(xml);
         }
         else {
            acFree(xml);
            DPrintF("!Program:","Failed to load menu file \"%s\"", Self->Menu);
         }

         FreeMemory(str);
      }
      else DPrintF("@Init:","[Window] Window specification does not support menus.");
   }

   if ((Self->ResizeFlags) AND (Self->ResizeBorder > 0) AND (Self->Screen IS FALSE) AND (Self->Drawable->ParentID)) {
      CreateObject(ID_RESIZE, NULL, &Self->Resize, NULL,
         FID_Object|TLONG,     Self->DrawableID,
         FID_Drawable|TLONG,   Self->DrawableID,
         FID_BorderSize|TLONG, Self->ResizeBorder,
         FID_Border|TLONG,     Self->ResizeFlags,
         TAGEND);
   }

   /* If the window is a screen, set it as the new default screen for opening
   ** further windows (with respect to this task).
   */

   if (Self->Screen IS TRUE) {
      glDefaultScreen = Self->DrawableID;
   }

   /*** If we are running in a hosted environment, set the window titlebar ***/

   if (!Self->Drawable->ParentID) {
      OBJECTID screen_id;
      OBJECTPTR screen;
      if (GetField(Self->Drawable, FID_Screen, FT_LONG, &screen_id) IS ERR_Okay) {
         if (AccessObject(screen_id, 4000, &screen) IS ERR_Okay) {
            SetField(screen, FID_Title, FT_STRING, Self->Title);
            ReleaseObject(screen);
         }
      }
   }

   if ((Self->Screen IS TRUE) OR (!Self->Drawable->ParentID)) {
      if (Self->Drawable->LeftMargin < Self->ScreenMargin)   Self->Drawable->LeftMargin   = Self->ScreenMargin;
      if (Self->Drawable->RightMargin < Self->ScreenMargin)  Self->Drawable->RightMargin  = Self->ScreenMargin;
      if (Self->Drawable->BottomMargin < Self->ScreenMargin) Self->Drawable->BottomMargin = Self->ScreenMargin;
   }

   SmartLimits(Self);

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  Maximise
Short: Maximises the window to its available display area.

###DESCRIPTION###
<p>Executing the Maximise method will run the maximise procedure for the
window. The specifics of the maximisation process may be environment
specific, but typically maximisation will result in the window being
expanded to the maximium dimensions allowed by its container.  The process
will take into account the hints provided by the margin settings of the
parent drawable (so if the margins are all set to 10, the window will be maximised
to the size of the container minus 10 pixels from its edge).</p>

<p>An optional Toggle argument allows you restore the window to its original
position if the method routine discovers that the window is already maximised.</p>

###INPUT###
LONG/Toggle: Set to TRUE to toggle the window back to its original dimensions if the window is already maximised.

###RESULT###
ERR_Okay: The window was maximised successfully.
ERR_AccessObject: Failed to access the window's parent drawable.

###END###

*****************************************************************************/

static ERROR WINDOW_Maximise(struct Window *Self, struct mtMaximiseWindow *Args)
{
   struct Render *parent;
   WORD x, y, insidewidth, insideheight;
   OBJECTPTR script;
   STRING str;

   if (Self->Maximise IS FALSE) return(ERR_Okay);

   /* The maximise script is run in addition to the normal maximise operation.
   ** If the developer wants a fully customised maximise operation, he can
   ** create his own custom widget in his window definition.
   */

   if ((str = GetXMLTemplateString(Self, "maximise"))) {
      if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
         FID_String|TSTRING, str,
         FID_Target|TLONG,   Self->DrawableID,
         FID_Window|TUNLISTED|TLONG, Self->Head.UniqueID,
         TAGEND) IS ERR_Okay) {

         acActivate(script);
         acFree(script);
      }

      FreeMemory(str);
   }

   if (AccessObject(Self->Drawable->ParentID, 5000, (OBJECTPTR *)&parent) IS ERR_Okay) {
      if ((x = parent->LeftMargin) < 0) x = 0;
      if ((y = parent->TopMargin) < 0) y = 0;
      insidewidth  = parent->Width  - parent->LeftMargin - parent->RightMargin;
      insideheight = parent->Height - parent->TopMargin  - parent->BottomMargin;

      ReleaseObject(parent);

      if ((Args) AND (Args->Toggle IS TRUE)) {
         DPrintF("Maximise()","[Window] Toggle-check.");

         /*** If the window is already maximised, restore it ***/

         if ((Self->RestoreWidth) AND (Self->RestoreHeight)) {
            if ((Self->Drawable->XCoord IS x) AND (Self->Drawable->YCoord IS y) AND
                (Self->Drawable->Width IS insidewidth) AND (Self->Drawable->Height IS insideheight)) {

               DPrintF("Maximise()","[Window] Restoring the window area.");

               ActionTags(AC_Redimension, Self->Drawable, (DOUBLE)Self->RestoreX, (DOUBLE)Self->RestoreY, 0.0,
                  (DOUBLE)Self->RestoreWidth, (DOUBLE)Self->RestoreHeight, 0.0);

               return(ERR_Okay);
            }
         }
      }

      if ((insidewidth IS Self->Drawable->Width) AND (insideheight IS Self->Drawable->Height)) {
         /* If the window is already at the required width and height,
         ** simply move the window rather than going through with the
         ** maximise process.
         */

         ActionTags(AC_MoveToPoint, Self->Drawable, (DOUBLE)x, (DOUBLE)y, 0.0, MTF_X|MTF_Y);
      }
      else {
         /*** Save current values ***/

         Self->RestoreX = Self->Drawable->XCoord;
         Self->RestoreY = Self->Drawable->YCoord;
         Self->RestoreWidth  = Self->Drawable->Width;
         Self->RestoreHeight = Self->Drawable->Height;

         DPrintF("Maximise()","[Window] Maximising the window area.");

         ActionTags(AC_Redimension, Self->Drawable, (DOUBLE)x, (DOUBLE)y, 0.0,
            (DOUBLE)insidewidth, (DOUBLE)insideheight, 0.0);
      }
   }
   else return(ObjectError(Self, MT_MaximiseWindow, ERR_AccessObject));

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  Minimise
Short: Minimises the size of the window.

###DESCRIPTION###
<p>Executing the Minimise method will run the minimise procedure for the
window. The specifics of the minimisation process are defined by the
window template of the current environment.  It is typical for the window to
iconify itself to the desktop in some way, so that it is removed from the
sight of the user.</p>

###RESULT###
ERR_Okay: Success.

###END###

*****************************************************************************/

static ERROR WINDOW_Minimise(struct Window *Self, APTR Void)
{
   OBJECTPTR script;
   STRING str;

   if (Self->Minimise IS FALSE) return(ERR_Okay);

   //ProcessMessages(NULL, NULL);

   DPrintF("~Minimise()","[Window:%d]", Self->Head.UniqueID);

   /*** Run the minimise script ***/

   if ((str = GetXMLTemplateString(Self, "minimise"))) {
      if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
         FID_String|TSTRING, str,
         FID_Target|TLONG,   Self->DrawableID,
         FID_Window|TUNLISTED|TLONG, Self->Head.UniqueID,
         TAGEND) IS ERR_Okay) {

         acActivate(script);
         acFree(script);
      }

      FreeMemory(str);
      StepBack();
      return(ERR_Okay);
   }
   else DPrintF("@Minimise:","No template script for minimisation is available.");

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  MoveToBack
Short: Moves the window to the back of the display area.
###END###

*****************************************************************************/

static ERROR WINDOW_Move(struct Window *Self, struct acMove *Args)
{
   return(Action(AC_Move, Self->Drawable, Args));
}

/*****************************************************************************

###ACTION###
Name:  MoveToBack
Short: Moves the window to the back of the display area.
###END###

*****************************************************************************/

static ERROR WINDOW_MoveToPoint(struct Window *Self, struct acMoveToPoint *Args)
{
   return(Action(AC_MoveToPoint, Self->Drawable, Args));
}

/*****************************************************************************

###ACTION###
Name:  MoveToBack
Short: Moves the window to the back of the display area.
###END###

*****************************************************************************/

static ERROR WINDOW_MoveToBack(struct Window *Self, APTR Void)
{
   return(Action(AC_MoveToBack, Self->Drawable, NULL));
}

/*****************************************************************************

###ACTION###
Name:  MoveToFront
Short: Moves the window to the front of the display area.
###END###

*****************************************************************************/

static ERROR WINDOW_MoveToFront(struct Window *Self, APTR Void)
{
   Action(AC_MoveToFront, Self->Drawable, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** Window: NewChild
*/

static ERROR WINDOW_NewChild(struct Window *Self, struct acNewChild *Args)
{
   OBJECTPTR newchild;

   if (!(Self->Head.ObjectFlags & NF_INITIALISED)) return(ERR_Okay);

   if (AccessObject(Args->NewChildID, 4000, &newchild) IS ERR_Okay) {
      SetOwner(newchild, Self->Drawable);
      ReleaseObject(newchild);
      return(ERR_OwnerPassThrough);
   }
   else {
      DPrintF("@NewChild:","[Window] Could not gain access to child object %d.", Args->NewChildID);
      return(ERR_AccessObject);
   }
}

/*****************************************************************************
** Window: NewObject()
*/

static ERROR WINDOW_NewObject(struct Window *Self, APTR Void)
{
   if ((!glDefaultScreen) OR (CheckObjectExists(glDefaultScreen, NULL) != ERR_Okay)) {
      /*** Set the default screen to the desktop if we haven't set it already **/
      FastFindObject("Desktop", NULL, &glDefaultScreen, 1, NULL);
   }

   if (NewObject(ID_RENDER, NF_CHILD, &Self->Drawable, &Self->DrawableID) IS ERR_Okay) {
      SetFields(Self->Drawable,
         FID_Name|TSTRING,    "winrender",
         FID_Parent|TLONG,    glDefaultScreen,
         FID_Width|TLONG,     300,
         FID_Height|TLONG,    300,
         FID_MinWidth|TLONG,  80,
         FID_MinHeight|TLONG, 40,
         FID_MaxWidth|TLONG,  4096,
         FID_MaxHeight|TLONG, 4096,
         TAGEND);

      Self->ProgramID    = Self->Head.TaskID;
      Self->ResizeBorder = 4;
      Self->Minimise     = TRUE;
      Self->Maximise     = TRUE;
      Self->MoveToBack   = TRUE;
      Self->Close        = TRUE;
      Self->Focus        = TRUE;
      Self->Quit         = TRUE;
      StrCopy("Window", Self->Title, sizeof(Self->Title));
      StrCopy("icons:devices/monitor", Self->Icon, sizeof(Self->Icon));

      SET_Template(Self, "templates:windowdef.xml");

      return(ERR_Okay);
   }
   else return(ERR_NewObject);
}

/*****************************************************************************
** Window: NewOwner()
*/

static ERROR WINDOW_NewOwner(struct Window *Self, struct acNewOwner *Args)
{
   /* If the new owner of the window is a drawable and we have not been
   ** through the initialisation process yet, switch to that drawable
   ** as our new window parent.
   */

   if (!(Self->Head.ObjectFlags & NF_INITIALISED)) {
      if (Args->ClassID IS ID_RENDER) {
         Self->Drawable->ParentID = Args->NewOwnerID;
      }
      else if (Args->ClassID IS ID_WINDOW) {
         struct Window *window;
         if (AccessObject(Args->NewOwnerID, 4000, (OBJECTPTR *)&window) IS ERR_Okay) {
            Self->Drawable->ParentID = window->DrawableID;
            ReleaseObject(window);
         }
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Redimension
Short: Changes the size and position of the window.
###END###

*****************************************************************************/

static ERROR WINDOW_Redimension(struct Window *Self, struct acRedimension *Args)
{
   return(Action(AC_Redimension, Self->Drawable, Args));
}

/*****************************************************************************
** Window: ReleaseObject
*/

static ERROR WINDOW_ReleaseObject(struct Window *Self, APTR Void)
{
   if (Self->Drawable) {
      ReleaseObject(Self->Drawable);
      Self->Drawable = NULL;
   }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Resize
Short: Alters the size of the window.
###END###

*****************************************************************************/

static ERROR WINDOW_Resize(struct Window *Self, struct acResize *Args)
{
   return(Action(AC_Resize, Self->Drawable, Args));
}

/*****************************************************************************

###ACTION###
Name:  SetUnlistedField
Short: Pass-through arguments can be set by using this action.
###END###

*****************************************************************************/

#define MAX_ARGS 40

static ERROR WINDOW_SetUnlistedField(struct Window *Self, struct acSetUnlistedField *Args)
{
   STRING str, Arg, field, value;
   MEMORYID ArgMID;
   WORD i;

   if (CurrentTaskID() != Self->Head.TaskID) return(ERR_Failed);

   /* Please note that it is okay to set zero-length string values
   ** (this has its uses in some scripts).
   */

   if ((!Args->Field) OR (!Args->Value)) return(ERR_Args);
   if (!Args->Field[0]) return(ERR_Args);

   field = Args->Field;
   value = Args->Value;

   if (!Self->LocalArgs) {
      if (AllocMemory(sizeof(STRING) * MAX_ARGS, NULL, (APTR *)&Self->LocalArgs, NULL) != ERR_Okay) {
         return(ObjectError(Self, ERH_SetUnlistedField, ERR_AllocMemory));
      }
      Self->TotalArgs = 0;
   }
   else if (Self->TotalArgs >= MAX_ARGS) return(ObjectError(Self, ERH_SetUnlistedField, ERR_ArrayFull));

   /* Check if the argument refers back to itself (e.g. 'location' = '{location}' would cause a loop-back)
   ** Set the field value to a null-string if such an occurance is detected.
   */

   if (value[0] IS '{') {
      UBYTE ch1, ch2;
      for (i=0; (field[i]) AND (value[i+1]); i++) {
         ch1 = field[i]; if ((ch1 >= 'A') AND (ch1 <= 'Z')) ch1 = ch1 - 'A' + 'a';
         ch2 = value[i+1];  if ((ch2 >= 'A') AND (ch2 <= 'Z')) ch2 = ch2 - 'A' + 'a';
         if (ch1 != ch2) break;
      }
      if ((!field[i]) AND (value[i+1] IS '}') AND (!value[i+2])) {
         DPrintF("@SetUnlistedField:","[Menu] Warning: Resetting looped argument '%s = %s'", field, value);
         value = "";
      }
   }

   if (AllocMemory(StrLength(field) + StrLength(value) + 2, MEM_STRING, (APTR *)&Arg, &ArgMID) IS ERR_Okay) {
      str = Arg;
      for (i=0; field[i]; i++) *str++ = field[i];
      *str++ = 0;

      for (i=0; value[i]; i++) *str++ = value[i];
      *str = 0;

      /*** Argument insertion ***/

      for (i=0; i < Self->TotalArgs; i++) {
         if (StrCompare(Arg, Self->LocalArgs[i], NULL, STR_MATCHLENGTH) IS ERR_Okay) {
            FreeMemory(Self->LocalArgs[i]);
            Self->LocalArgs[i] = Arg;
            break;
         }
      }

      if (i IS Self->TotalArgs) {
         Self->LocalArgs[Self->TotalArgs] = Arg;
         Self->TotalArgs++;
      }

      ReleaseMemory(Arg);
   }
   else return(ERR_Memory);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Show
Short: Puts the window on display.
###END###

*****************************************************************************/

static ERROR WINDOW_Show(struct Window *Self, APTR Void)
{
   OBJECTPTR object;
   OBJECTID object_id;

   DPrintF("~Show()","[Window:%d]", Self->Head.UniqueID);

   if (Self->Focus) {
      if (!(Self->Drawable->Flags & RNF_HASFOCUS)) {
         acFocus(Self->Drawable);
      }
   }

   acShow(Self->Drawable);

   if (CheckObjectExists(NULL, "SystemPointer") != ERR_Okay) {
      if (NewNamedObject(ID_POINTER, NF_NOTRACK|NF_PUBLIC|NF_UNIQUE,
         &object, &object_id, "SystemPointer") IS ERR_Okay) {
         SetField(object, FID_Drawable, FT_LONG, &Self->DrawableID);
         if (acInit(object) != ERR_Okay) acFree(object);
         acShow(object);
         ReleaseObject(object);
      }
   }

   if (CheckObjectExists(NULL, "SystemKeyboard") != ERR_Okay) {
      if (CreateObject(ID_KEYBOARD, NF_PUBLIC|NF_UNTRACKED|NF_UNIQUE, &object, &object_id,
            FID_Name|TSTRING, "SystemKeyboard",
            TAGEND) IS ERR_Okay) {
         acActivate(object);
         ReleaseObject(object);
      }
   }

   Self->Shown = TRUE;

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Canvas
Short:  Allocates a drawable canvas inside the window when read.
Type:   OBJECTID
Status: Get

###DESCRIPTION###
<p>To automatically allocate a drawable inside a window after it has been
initialised, read the Canvas field.  A simple drawable matching the internal
dimensions of the window will be created and its object ID will be returned
as the field value.</p>

<p>You can make adjustments to the drawable if desired, but if you have special
requirements then it is recommended that you create a suitable canvas manually.</p>

<p>The Canvas field can only be used following initialisation.  Once the window
has been shown on the display, the Canvas field will cease to operate in order to
prevent other tasks from accidentally creating canvasses when they read this
field.</p>

###END###

*****************************************************************************/

static ERROR GET_Canvas(struct Window *Self, OBJECTID *Value)
{
   OBJECTPTR drawable;

   if (!(Self->Head.ObjectFlags & NF_INITIALISED)) return(ObjectError(Self, ERH_GetField, ERR_NotInitialised));

   if ((Self->CanvasID) OR (Self->Shown)) {
      *Value = Self->CanvasID;
      return(ERR_Okay);
   }

   if (CreateObject(ID_RENDER, NULL, &drawable, &Self->CanvasID,
      FID_Name|TSTRING,    "winCanvas",
      FID_Parent|TLONG,    Self->Drawable->Head.UniqueID,
      FID_XCoord|TLONG,    Self->Drawable->LeftMargin,
      FID_YCoord|TLONG,    Self->Drawable->TopMargin,
      FID_XOffset|TLONG,   Self->Drawable->RightMargin,
      FID_YOffset|TLONG,   Self->Drawable->BottomMargin,
      TAGEND) IS ERR_Okay) {

      acShow(drawable);
      ReleaseObject(drawable);
      *Value = Self->CanvasID;
      return(ERR_Okay);
   }
   else return(ERR_CreateObject);
}

/*****************************************************************************

###FIELD###
Name:   Center
Short:  Displays the window in the center of the display on initialisation.
Type:   BOOLEAN
Status: Read/Init

###DESCRIPTION###
<p>The Center field may be set to TRUE on initialisation for the window to
be automatically centered before it is shown for the first time.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Close
Short:  Operator for the window's close gadget.
Type:   BOOLEAN
Status: Read/Set

###DESCRIPTION###
<p>The Close field controls the activity of the close gadget, which is typically
found in the window title bar.  The default value for the close gadget is TRUE
and may be switched off at any time by setting this field to FALSE.</p>

###END###

*****************************************************************************/

static ERROR SET_Close(struct Window *Self, LONG *Value)
{
   if (*Value IS TRUE) {
      Self->Close = TRUE;
      if ((Self->Head.ObjectFlags & NF_INITIALISED) AND (Self->CloseID)) {
         ActionMsg(AC_Enable, Self->CloseID, NULL);
      }
   }
   else {
      Self->Close = FALSE;
      if ((Self->Head.ObjectFlags & NF_INITIALISED) AND (Self->CloseID)) {
         ActionMsg(AC_Disable, Self->CloseID, NULL);
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   CloseObject
Short:  The drawable that controls the Close gadget.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>The drawable that represents the window's close gadget may be referenced
through this field.  If no close gadget is available then this field will be
empty.</p>

<p>Theme Developers: Your window code must set this field with the ID of the
drawable that you have created for the close gadget.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  The window's drawable.
Type:   struct Render *
Status: Read

###DESCRIPTION###
<p>The drawable that represents the window on the display can be retrieved
through this field.  The reference is a direct pointer to the Render object
in question. It is possible to manipulate the drawable without
encountering adverse effects with the window object, although it is wise to
refrain from doing so where functionality is duplicated in the Window class.
For instance, Hide and Show actions may not be called on the drawable object
as this functionality is already controlled by the window system.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Optional flags may be set here.
Type:   LONG
Prefix: WNF
Status: Read/Init

###DESCRIPTION###
<p>Optional flags that may be set against a window object are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">SMARTLIMITS</td><td>Enables smart-limit settings, a special feature that prevents the window from completely leaving the display no matter where the user tries to move the window.  This mode will ensure that enough of the title bar is visible so as to allow the user to drag the window back into the display area.</td></tr>
<tr><td width="10%">MANUALCLOSE</td><td>This option is used for controlling the behaviour taken when the window is asked to close.  You must match this flag with a subscription to the Close method so that you can respond to window close signals.</td></tr>
<tr><td width="10%">BACKGROUND</td><td>This option will guarantee that the background of the window will always be cleared to a neutral colour (probably grey).  Background clearing occurs automatically for desktop windows, but not for windows that are in full-screen mode.  You should always set this flag if you do not intend to define the window background yourself.</td></tr>
<tr><td width="10%">VIDEO</td><td>Enables video buffering of the the window instead of software buffering.  The success of this option is dependent on whether or not video RAM is available.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Focus
Short:  May be set to FALSE if you wish to open the window without giving it the focus.
Type:   BOOLEAN
Status: Read/Write

###DESCRIPTION###
<p>By default the window will be given the focus whenever it is shown on the
display via the Show action.  You turn off this behaviour and take control
of the window focus manually by setting the Focus field to FALSE.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Hide
Short:  Set on initialisation to keep the window hidden.
Type:   BOOLEAN
Status: Read/Init

###DESCRIPTION###
<p>This field is for use when creating windows in scripts, such as
DML.  By default, windows are shown following creation in scripts.  You can
turn off this behaviour and manually show the window when you are ready by
setting the Hide value to TRUE.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Icon
Short:  A graphical icon to be applied to the window.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>A suitable icon that you would like to apply to the window may be specified
here.  The window theme will choose how the icon is used and displayed to the
user.  It is common for windows to use the referenced icon in the window title
bar and when iconifying the window.</p>

<p>It is preferred that you use an icon from the existing icon database (if
you have a customised icon that you would like to use, you should install
it to the icon database on installation).  The window system will automatically
scale the icon graphic to the required dimensions when it is loaded.</p>

###END###

*****************************************************************************/

static ERROR GET_Icon(struct Window *Self, STRING *Value)
{
   if (Self->Icon) {
      *Value = Self->Icon;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_Icon(struct Window *Self, STRING Value)
{
   WORD i;

   if ((!Value) OR (!Value[0])) return(ERR_Failed);

   if (StrCompare("icons:", Value, 0, NULL) IS ERR_Okay) {
      for (i=0; (i < sizeof(Self->Icon)-1) AND (Value[i]); i++) Self->Icon[i] = Value[i];
      Self->Icon[i] = 0;
   }
   else {
      i = StrCopy("icons:", Self->Icon, sizeof(Self->Icon));
      StrCopy(Value, Self->Icon + i, sizeof(Self->Icon)-i);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   InsideBorder
Short:  Enables a custom graphics border that surrounds the internal window.
Type:   BOOLEAN
Status: Read/Init

###DESCRIPTION###
<p>Set the InsideBorder field to TRUE to create a simple graphics border
around the edges of the drawable window area.  The benefit of using this field
to create a graphics border instead of a customised border is that it maintains
consistency with the other window graphics for the loaded environment.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Maximise
Short:  Operator for the window's maximise gadget.
Type:   BOOLEAN
Status: Read/Set

###DESCRIPTION###
<p>The Maximise field controls the activity of the maximise gadget, which is
typically found in the window title bar.  The default value for the gadget
is TRUE (environment dependent) and may be switched off at any time by
setting this field to FALSE.</p>

###END###

*****************************************************************************/

static ERROR SET_Maximise(struct Window *Self, LONG *Value)
{
   if (*Value IS TRUE) {
      Self->Maximise = TRUE;
      if ((Self->Head.ObjectFlags & NF_INITIALISED) AND (Self->MaximiseID)) {
         ActionMsg(AC_Enable, Self->MaximiseID, NULL);
      }
   }
   else {
      Self->Maximise = FALSE;
      if ((Self->Head.ObjectFlags & NF_INITIALISED) AND (Self->MaximiseID)) {
         ActionMsg(AC_Disable, Self->MaximiseID, NULL);
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   MaximiseObject
Short:  The drawable that controls the maximise gadget.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>The drawable that represents the window's maximise gadget may be
referenced through this field.  If no maximise gadget is available then this
field will be empty.</p>

<p>Theme Developers: Your window code must set this field with the ID of the
drawable that you have created for the maximise gadget.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Menu
Short:  References the location of an XML menu file to be applied to the window.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>To load a menu bar into the window you will need to set this field with
the location of a menu file that you want to use for the window.  The file
must be in XML format and contain at least one menu tag.  Each menu tag that
you specify will create a new menu option in the menu bar that is owned by the
window.  Non-menu tags will be ignored.</p>

<p>Any unlisted arguments that you have specified for the window will be
passed on to each created menu.</p>

<p>For information on the specifics of menu customisation, please
refer to the Menu class.</p>

###SEE ALSO###
Class: Menu

###END###

*****************************************************************************/

static ERROR GET_Menu(struct Window *Self, STRING *Value)
{
   if (Self->Menu) {
      *Value = Self->Menu;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_Menu(struct Window *Self, STRING Value)
{
   StrCopy(Value, Self->Menu, sizeof(Self->Menu));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   MenubarObject
Short:  The drawable that controls the window's menu bar.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>If you have created a menu for the window then this field will reference
the menu bar's drawable area.</p>

<p>Theme Developers: Your menu bar creation code must set this field with
the ID of the drawable that you have created for menu bar
representation.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Minimise
Short:  Operator for the title bar's minimise gadget.
Type:   BOOLEAN
Status: Read/Init

###DESCRIPTION###
<p>The Minimise field controls the activity of the minimise gadget, which is
typically found in the window title bar.  The default value for the gadget
is TRUE (environment dependent) and may be switched off at any time by
setting this field to FALSE.</p>

###END###

*****************************************************************************/

static ERROR SET_Minimise(struct Window *Self, LONG *Value)
{
   if (*Value IS TRUE) {
      Self->Minimise = TRUE;
      if ((Self->Head.ObjectFlags & NF_INITIALISED) AND (Self->MinimiseID)) {
         ActionMsg(AC_Enable, Self->MinimiseID, NULL);
      }
   }
   else {
      Self->Minimise = FALSE;
      if ((Self->Head.ObjectFlags & NF_INITIALISED) AND (Self->MinimiseID)) {
         ActionMsg(AC_Disable, Self->MinimiseID, NULL);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   MinimiseObject
Short:  The drawable that controls the minimise gadget.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   MoveToBack
Short:  Operator for the title bar's move-to-back gadget.
Type:   BOOLEAN
Status: Read/Init

###DESCRIPTION###
<p>The MoveToBack field controls the activity of the move-to-back gadget,
which is typically found in the window title bar.  The default value for the
gadget is TRUE (environment dependent) and may be switched off at any time
by setting this field to FALSE.</p>

###END###

*****************************************************************************/

static ERROR SET_MoveToBack(struct Window *Self, LONG *Value)
{
   if (*Value IS TRUE) {
      Self->MoveToBack = TRUE;
      if ((Self->Head.ObjectFlags & NF_INITIALISED) AND (Self->MoveToBackID)) {
         ActionMsg(AC_Enable, Self->MoveToBackID, NULL);
      }
   }
   else {
      Self->MoveToBack = FALSE;
      if ((Self->Head.ObjectFlags & NF_INITIALISED) AND (Self->MoveToBackID)) {
         ActionMsg(AC_Disable, Self->MoveToBackID, NULL);
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   MoveToBackObject
Short:  The drawable that controls the move-to-back gadget.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>The drawable that represents the window's move-to-back gadget may be
referenced through this field.  If no move-to-back gadget is available then this
field will be empty.</p>

<p>Theme Developers: Your window code must set this field with the ID of the
drawable that you have created for the move-to-back gadget.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Quit
Short:  Set to FALSE to prevent application termination when the window is closed.
Type:   BOOLEAN
Status: Read/Write

###DESCRIPTION###
<p>By default this field is set to TRUE, which will force the program to quit
when the window is closed (the application will be sent a QUIT message).
You can turn off this behaviour by setting this flag to FALSE.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Resize
Short:  Determines what sides of the window are resizable.
Type:   STRING
Status: Init

###DESCRIPTION###
<p>This field defines what sides of the window are resizable.  It can only
be set from the template script that is defined for the window.  The string
format is defined as a series of flags separated with the OR character.
Valid options are:</p>

<pre>
   TOP, BOTTOM, LEFT, RIGHT
   TOPLEFT, TOPRIGHT, BOTTOMLEFT, BOTTOMRIGHT
</pre>

###SEE ALSO###
Field: ResizeBorder

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ResizeBorder
Short:  Defines the extent of the resize area at the sides of the window.
Type:   LONG
Status: Init

###DESCRIPTION###
<p>If you have set the Resize field then you can optionally define the size
of the border to use for each window edge.  Recommended values range between
4 and 8 pixels.</p>

<p>This field can only be set from the template script that is defined for
the window.</p>

###SEE ALSO###
Field: Resize

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   RestoreHeight
Short:  Controls the height of the window when restoring to its previous state.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field assists in the restoration of the window when it is minimised,
maximised or has its dimensions changed in a customised way.  The value that
you set here will reflect the new height of the window when it is restored
to its 'previous' state.  This field is provided for control purposes and
should only be used from the window's template script.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   RestoreWidth
Short:  Controls the width of the window when restoring to its previous state.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field assists in the restoration of the window when it is minimised,
maximised or has its dimensions changed in a customised way.  The value that
you set here will reflect the new width of the window when it is restored
to its 'previous' state.  This field is provided for control purposes and
should only be used from the window's template script.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   RestoreX
Short:  Controls the horizontal position of the window when restoring to its previous state.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field assists in the restoration of the window when it is minimised,
maximised or has its dimensions changed in a customised way.  The value that
you set here will reflect the new horizontal coordinate of the window when
it is restored to its 'previous' state.  This field is provided for control
purposes and should only be used from the window's template script.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   RestoreY
Short:  Controls the vertical position of the window when restoring to its previous state.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field assists in the restoration of the window when it is minimised,
maximised or has its dimensions changed in a customised way.  The value that
you set here will reflect the new vertical coordinate of the window when it
is restored to its 'previous' state.  This field is provided for control
purposes and should only be used from the window's template script.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Screen
Short:  Enables full-screen mode.
Type:   BOOLEAN
Status: Get/Init

###DESCRIPTION###
<p>Set the Screen field to TRUE in order to launch the window in full-screen
mode when it is opened.  Full-screen windows use a different window template
to that used as standard and thus have their own behavioural conditions that
may differ significantly to the default.  At a minimum, full-screen windows
will cover the entire display area and be locked to that size.  Resizing of
the window is not possible unless the user changes the desktop's screen
resolution.</p>

<p>When your application opens a full-screen window, it will serve as the
default display area when you open further windows.  If you would like to open
new windows in other drawables, you will need to set the Target field
to alter the default opening behaviour.</p>

###SEE ALSO###
Field: ScreenMargin

###END###

*****************************************************************************/

static ERROR SET_Screen(struct Window *Self, LONG *Value)
{
   if (*Value IS TRUE) {
      Self->Screen = TRUE;
      Self->Center = FALSE;
      SET_Template(Self, "templates:screendef.xml");
   }
   else Self->Screen = FALSE;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ScreenMargin
Short:  Defines the minimum amount of margin space if the window is in full-screen mode.
Type:   LONG
Status: Get/Init

###DESCRIPTION###
<p>If a window opens in full-screen mode (i.e. if you set the Screen field
to TRUE) then it will typically be created without any margins on the left,
right and bottom sides of the window space.  This will mean that all created
graphics will be flush against the screen area.  If you would like a small
amount of white-space to be added to the screen border then you will need to
set the ScreenMargin field.  A value between 4 and 6 is usually
appropriate.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ScriptArgs
Short:  Readable field for template usage only, retrieves custom arguments in DML format.
Type:   STRING
Status: Get

###DESCRIPTION###
<p>For template usage only.  This field will return a string that contains
all unlisted arguments as a DML string series.  For instance, if you pass
two unlisted arguments 'A' and 'B' with values of '1' and '0' respectively,
reading the ScriptArgs argument will return the following:</p>

<pre>
   &A="1" &B="0"
</pre>

<p>This feature is typically used in the template definition file for
passing custom parameters to objects such as menus.</p>

###END###

*****************************************************************************/

static ERROR GET_ScriptArgs(struct Window *Self, STRING *Value)
{
   #define STRLEN 4096
   OBJECTPTR context;
   STRING arg, str, value;
   ERROR error;
   WORD i, pos;

   if (CurrentTaskID() != Self->Head.TaskID) return(ERR_Failed);

   if (Self->ScriptArgs) {
      FreeMemory(Self->ScriptArgs);
      Self->ScriptArgs = NULL;
   }

   context = SetContext(Self);
   error = AllocMemory(STRLEN, MEM_STRING|MEM_NOCLEAR, &Self->ScriptArgs, NULL);
   SetContext(context);

   if (error IS ERR_Okay) {
      str = Self->ScriptArgs;
      pos = 0;
      for (i=0; i < Self->TotalArgs; i++) {
         if ((arg = Self->LocalArgs[i])) {
            value = arg;
            while (*value) value++;
            value++;

            StrFormat(str+pos, STRLEN-pos, "&%s=\"$%s$\" ", arg, value);
            while (str[pos]) pos++;

            arg = value + 1;
         }
      }
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AllocMemory));

   *Value = Self->ScriptArgs;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   StickToFront
Short:  Forces the window to stick to the front of the display.
Type:   BOOLEAN
Status: Set

###DESCRIPTION###
<p>Set this field to TRUE to force the window to stay at the front of its
display.  This option should be used leniently.  It is recommended for use
in small windows and presenting warning or error messages to the user.</p>

###END###

*****************************************************************************/

static ERROR SET_StickToFront(struct Window *Self, LONG *Value)
{
   SetField(Self->Drawable, FID_StickToFront, FT_LONG, Value);
   if (*Value IS TRUE) Self->MoveToBack = FALSE;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Template
Short:  Defines the makeup of the window using a pre-defined template.
Type:   STRING
Status: Init

###DESCRIPTION###
<p>Setting the Template field allows a window definition to be loaded from a
pre-defined template file.  Changing to a customised template is not
recommended, as the user environment will have already defined a preferred
template file for you.  Use of customised templates may cause problems with
consistency in the user interface.  If you would like to define a window template
for a new environment, we recommend defining a new templates:windowdef.xml file
rather than setting this field directly.</p>

<p>Template files are defined in XML and are split into a number of segments
that control specific parts of the window.  The first definable tag is
'values', which defines the default values for newly created windows.  It is
typical to define preferred values for the window coordinates and set the
resize parameters.  The window margins should also be configured so that it
is clear how much space around the edges of the window is for custom
graphics.  Here is an example:</p>

<pre>
&lt;values&gt;
  &lt;resize value="left|bottomleft|bottom|bottomright|right"/&gt;
  &lt;xcoord value="=[[self.parent].leftmargin]+10"/&gt;
  &lt;ycoord value="=[[self.parent].topmargin]+10"/&gt;
  &lt;resizeborder value="4"/&gt;
  &lt;topmargin value="22"/&gt;
  &lt;leftmargin value="6"/&gt;
  &lt;rightmargin value="6"/&gt;
  &lt;bottommargin value="6"/&gt;
&lt;/values&gt;
</pre>

<p>The bulk of the window graphics are defined in the 'graphics' tag.  You are
expected to create graphics objects that will be initialised directly to the
window's drawable.  The drawable can be referenced via the {owner} argument
and the window referenced through the {window} argument.  Your script should
also take note of the InsideBorder field and create a window border if the
border has a value of 1.  The following example illustrates:</p>

<pre>
&lt;graphics&gt;
  &lt;box colour="[gl_colours.colour]" shadow="[gl_colours.shadow]"
    highlight="[gl_colours.highlight]" raised/&gt;

  &lt;if statement="[{window}.insideborder] = 1"&gt;
    &lt;box sunken x="=[owner.leftmargin]-1" y="=[owner.topmargin]-1"
      xoffset="=[owner.rightmargin]-1" yoffset="=[owner.bottommargin]-1"
      highlight="#dfdfdf" shadow="#202020"/&gt;

    &lt;box sunken x="=[owner.leftmargin]-2" y="=[owner.topmargin]-2"
      xoffset="=[owner.rightmargin]-2" yoffset="=[owner.bottommargin]-2"
      highlight="#ffffff" shadow="#808080"/&gt;
  &lt;/if&gt;
&lt;/graphics&gt;
</pre>

<p>The window title bar is configured in the 'titlebar' tag.  The
title bar should allow the user to drag the window around the display
and provide a series of gadgets that control the window (such as minimise
and maximise buttons).  You should also define an object that will
control the window title.  You will need to communicate with the window
object for the purpose of determining the titlebar configuration, plus
you will need to set certain window fields for communication purposes.
For examples on how to generate the titlebar, please refer to your
templates:windowdef.xml file.</p>

<p>You will need to define a 'menubar' tag so that a menu strip can be
created for the window if the user requests one.  The requirements for
this are similar to the title bar definition.  An {xml} parameter will
be provided that refers to an XML object that is to be used for configuring
the menu bar.  By scanning the object you will be able to extract the
names of each menu option that you need to create.  Please refer to your
templates:windowdef.xml file to observe how this should be done.</p>

<p>Finally, a number of optional tags may be used for window control purposes.
These tags can contain DML scripts that will be executed when certain actions
occur to the window.  The following table describes the available tags:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Tag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">Maximise</td><td>Executed when the Maximise method is called on the window.</td></tr>
<tr><td width="10%">Minimise</td><td>Executed when the Minimise method is called on the window.</td></tr>
</table>

###END###

*****************************************************************************/

static ERROR SET_Template(struct Window *Self, STRING Value)
{
   struct XMLTag *tag, *child;
   UBYTE value[300], calc[300];
   WORD i;

   if (Self->Head.ObjectFlags & NF_INITIALISED) return(ERR_Failed);

   if (Self->TemplateXML) {
      Action(AC_Free, Self->TemplateXML, NULL);
      Self->TemplateXML = NULL;
   }

   if (AnalyseLocation(Value, NULL) IS ERR_Okay) {
      if (CreateObject(ID_XML, NF_CHILD, &Self->TemplateXML, NULL,
         FID_Location|TSTRING, Value,
         TAGEND) IS ERR_Okay) {

         if ((tag = GetXMLTemplate(Self, "values"))) {
            for (child=tag->Child; child; child=child->Next) {
               if (child->TotalAttributes > 1) {
                  for (i=0; (i < sizeof(value)-1) AND (child->Attributes[1].Value[i]); i++) value[i] = child->Attributes[1].Value[i];
                  value[i] = 0;
                  if (value[0] IS '=') {
                     StrTranslate(value, sizeof(value), NULL);
                     if (StrCalculate(value+1, NULL, calc, sizeof(calc)) IS ERR_Okay) {
                        SetFieldVariable(Self, child->Attributes[0].Name, calc);
                     }
                     else SetFieldVariable(Self, child->Attributes[0].Name, value);
                  }
                  else {
                     StrTranslate(value, sizeof(value), NULL);
                     SetFieldVariable(Self, child->Attributes[0].Name, value);
                  }
               }
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Title
Short:  The title string to display in the window's title bar.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The Title field controls the text that is displayed in the window's title
bar.  You may set this field at any time in order to change the title that
is displayed inside the window.  A standard UTF-8 text string is acceptable
and will be trimmed if it exceeds the maximum number of possible characters.</p>

###END###

*****************************************************************************/

static ERROR GET_Title(struct Window *Self, STRING *Value)
{
   if (Self->Title) {
      *Value = Self->Title;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_Title(struct Window *Self, STRING Value)
{
   OBJECTPTR title;
   WORD i;

   DPrintF("Set.Title:","[Window] = %s", Value);

   if (Value) {
      for (i=0; (i < sizeof(Self->Title)-1) AND (Value[i]); i++) Self->Title[i] = Value[i];
      Self->Title[i] = 0;
   }
   else Self->Title[0] = 0;

   if (Self->TitleID) {
      if (AccessObject(Self->TitleID, 5000, &title) IS ERR_Okay) {
         SetField(title, FID_String, FT_STRING, Self->Title);
         ReleaseObject(title);
      }
   }

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      if (!Self->Drawable->ParentID) {
         OBJECTID screen_id;
         OBJECTPTR screen;
         if (GetField(Self->Drawable, FID_Screen, FT_LONG, &screen_id) IS ERR_Okay) {
            if (AccessObject(screen_id, 4000, &screen) IS ERR_Okay) {
               SetField(screen, FID_Title, FT_STRING, Self->Title);
               ReleaseObject(screen);
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   TitleObject
Short:  The object that controls the window title is referenced here.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>The object that controls the window's title bar text is referenced through
this field.  If no title bar is configured then this field will be empty.</p>

<p>Theme Developers: Your window code must set this field with the ID of the
object that manages the title bar.  A Text object is typically used for this
purpose.</p>

###END###

*****************************************************************************/

/*****************************************************************************
** Internal: CheckWindowOverlap()
*/

static void CheckWindowOverlap(struct Window *Self)
{
   struct RenderList *list, *renderlist;
   struct acMoveToPoint moveto;
   LONG x, y, size;

   if (Self->Screen IS FALSE) {
      if ((renderlist = AccessRenderList(ARF_READ, &size))) {
         x = Self->Drawable->XCoord;
         y = Self->Drawable->YCoord;
         for (list=renderlist; list->ObjectID; list=((APTR)list)+size) {
            if (list->ParentID IS Self->Drawable->ParentID) {
               if ((list->XCoord IS x) AND (list->YCoord IS y)) {
                  x += 20;
                  y += 20;
                  list = renderlist;
               }
            }
         }

         if ((x != Self->Drawable->XCoord) OR (y != Self->Drawable->YCoord)) {
            for (list=renderlist; list->ObjectID; list=((APTR)list)+size) {
               if (list->ObjectID IS Self->Drawable->ParentID) break;
            }

            if ((x + Self->Drawable->Width < list->Width) AND (y + Self->Drawable->Height < list->Height)) {
               moveto.XCoord = x;
               moveto.YCoord = y;
               moveto.Flags  = MTF_X|MTF_Y;
               Action(AC_MoveToPoint, Self->Drawable, &moveto);
            }
         }

         ReleaseRenderList(ARF_READ);
      }
   }
}

/*****************************************************************************
** Internal: GetXMLTemplate()
*/

static struct XMLTag * GetXMLTemplate(struct Window *Self, STRING String)
{
   struct XMLTag **xmllist, *Tag;

   if (GetFields(Self->TemplateXML, FID_Tags|TPTR, &xmllist, TAGEND) IS ERR_Okay) {
      Tag = xmllist[0];
      while (Tag) {
         if (Tag->Child) {
            if (StrCompare(String, Tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               return(Tag);
            }
         }
         Tag = Tag->Next;
      }
   }
   return(NULL);
}

/*****************************************************************************
** Internal: GetXMLTemplateString()
*/

static STRING GetXMLTemplateString(struct Window *Self, STRING String)
{
   struct XMLTag *tag;
   struct mtGetXMLString getxml;

   if ((tag = GetXMLTemplate(Self, String))) {
      getxml.Index = tag->Child->Index;
      getxml.Flags = XMF_INCLUDESIBLINGS;
      if (Action(MT_GetXMLString, Self->TemplateXML, &getxml) IS ERR_Okay) {
         return(getxml.Result);
      }
   }
   return(NULL);
}

/*****************************************************************************
** Internal: ProcessArguments()
*/

static void ProcessArguments(struct Window *Self, OBJECTPTR Script)
{
   struct acSetUnlistedField set;
   STRING arg;
   WORD i;

   for (i=0; i < Self->TotalArgs; i++) {
      if ((arg = Self->LocalArgs[i])) {
         set.Field = arg;
         while (*arg) arg++;
         arg++;
         set.Value = arg;
         Action(AC_SetUnlistedField, Script, &set);
      }
   }
}

/*****************************************************************************
** Internal: SmartLimits()
*/

static void SmartLimits(struct Window *Self)
{
   struct RenderInfo info;

   if ((Self->Screen IS FALSE) AND (Self->Flags & WNF_SMARTLIMITS) AND (Self->Drawable->ParentID)) {
      if (GetRenderInfo(Self->Drawable->ParentID, &info, sizeof(info)) IS ERR_Okay) {
         Self->Drawable->TopLimit    = 0;
         Self->Drawable->BottomLimit = -Self->Drawable->Height + Self->Drawable->TopMargin;
         Self->Drawable->LeftLimit   = -(Self->Drawable->Width * 0.75);
         Self->Drawable->RightLimit  = -(Self->Drawable->Width * 0.75);
      }
   }
}

/*****************************************************************************
** Internal: set_variables()
*/

static void set_variables(OBJECTPTR Variable, struct VarString *Fields)
{
   WORD i;
   for (i=0; Fields[i].Field; i++) {
      SetUnlistedField(Variable, Fields[i].Field, Fields[i].Value);
   }
}
